#ifndef CLI_H
#define CLI_H
#include <stdlib.h>

struct cli_blok;

typedef int parsefun_t (const char * /* arg */ ,
			struct cli_blok * /* k čemu se to vztahuje */ ,
			struct cli_blok * /* Nový blok, nebo NULL */ );

struct cli_opt
{
  char *name;			// Musí být unikátní. Bez - na začátku
  int mandatory;		// Minimální počet výskytů
  int has_arg;			// Nemá-li argument, 0, jinak 1
  parsefun_t *parsefun;		// Nemá-li co na práci (pouze nový blok), patří zde NULL
  const struct cli_blok **choices;	// Vyplnit šablony, volí-li se blok. Ukončeno NULL. Jinak NULL.
  /* Při deklaraci nastavit výchozí hodnoty pomocí makra OPDEF: */
  int count;
  struct cli_blok *blok;	// Blok, jemuž tato volba přísluší
};
#define OPDEF 0 , NULL
#define OPT_EMPTY {NULL, 0, 0, NULL, NULL, 0, NULL}

struct cli_blok
{
  char *name;
  struct cli_opt *optlist;	// Pole voleb pro blok.
  //TODO odstranit opt_count a počítat to podle OPT_EMPTY na konci optlist
  int opt_count;		// Počet voleb pro blok (v. předchozí řádek)
  void (*fn) ();		// Funkce odpovídající tomuto bloku (např. cold pro blok cold)
  size_t data_sz;		// Velikost datové oblasti
  int (*checkfn) (void *);	// Funkce, která se volá na data po načtení všech voleb
  void (*data_clean) (void *);	// „Destruktor“ obsahu dat (avšak bez free(data)!)
  void *data;			// Datová oblast bloku
};

/* TODO const peklo */
typedef const struct cli_blok *(choices_t[]);
typedef struct cli_opt optlist_t[];

/* 
 * Najde volbu cli_opt se zadaným jménem v poli (ukončeném OPT_EMPTY).
 * Vrací první nalezenou volbu s přesně odpovídajícím jménem.
 * Nenajde-li, vrací NULL.
 */
struct cli_opt *findopt (struct cli_opt *array, const char *name);

/* 
 * Najde volbu cli_opt začínající na daný řetězec v poli (ukončeném OPT_EMPTY). 
 * Nenajde-li, vrací NULL. V případě víceznačnosti vypíše chybu na stderr
 * a vrací NULL.
 */
struct cli_opt *pfindopt (struct cli_opt *array, const char *name);

/*
 * Totéž s poli ukazatelů na cli_opt (ukončeném NULL).
 */
struct cli_opt *findpopt (struct cli_opt **ptrarray, const char *name);
struct cli_opt *pfindpopt (struct cli_opt **ptrarray, const char *name);

/*
 * Podobné, ale pro bloky (avšak s ukazateli!)
 */
const struct cli_blok *findblok (const struct cli_blok **array,
				 const char *name);
const struct cli_blok *pfindblok (const struct cli_blok **array,
				  const char *name);

//viz cli_misc.c
#include "matvec.h"
void FATAL (const char *fmt, ...);
void WARN (const char *fmt, ...);
int parse_vec (const char *i, rvec * o);
int parse_doubles (const char *in, double o[], int maxcount);
int parse_longs (const char *in, long int o[], int maxcount);
int iszero (const void *data, size_t nbytes);
void printchoices (const struct cli_opt *o);	// Vypíše názvy bloků na stderr

/* TODO?  ==== Níže uvedené by snad klidně mohlo zůstat v cli_templates.c ==== */
// viz cli_templates.c a cli.c

#include "trace.h"
struct trace1_p
{				// Viz příslušné parametry pro trace() v trace.h
  rvec from;
  rvec direction;
  int (*stop_condition) (rvec r, void *param);
  void *stop_condition_p;
  double frequency;
  double maxt;
  double step;
  susc_t *susc;
  void *susc_p;
  int mod;
};

// Datová struktura pro parse_density_profile
#include "cold.h"		// TODO? zobecnit densf_t šířeji než do cold.h
struct density_profile_p
{
  densf_t *densf;
  void *densf_p;
  double step;
  rvec from;
  rvec to;
};


// Datová struktura pro parse_magfield_profile
struct magfield_profile_p
{
  magf_t *magf;
  void *magf_p;
  double step;
  int stepcount;
  rvec from;
  rvec to;
};

// Datová struktura pro dispfun_view
struct dispfun_view_p
{
  rvec position;
  rvec direction;
  double frequency;
  susc_t *susc;
  void *susc_p;
  double from;
  double to;
  double step;
};

#endif //CLI_H
