#ifndef COLD_H
#define COLD_H
#include "trace.h"
/* 
 * Obecná susceptibilita studeného plasmatu: S libovolným
 * polem a s libovolným počtem složek
 */

/* Typ pro funkci vracející parametry plasmatu v závislosti na poloze *r:
 * do předalokovaného pole dest[] uloží hustoty složek v patřičném pořadí
 * s volitelnými parametry v param. */
typedef void densf_t (double *dest, rvec * r, void *param);


/* 
 * Parametry pro obecnou susceptibilitu studeného plasmatu
 * Není-li řečeno jinak, pole v parametrech si alokuje volající sám!
 */
struct cold_p
{
  int s;			// počet složek (vč. elektronů)
  double *m;			// hmotnost částice
  double *q;			// náboj částice
  densf_t *densf;
  void *densf_p;
  magf_t *magf;
  void *magf_p;
};

/* Nastavení "polních" parametrů */
void cold_p_set_s (struct cold_p *p, int news);
void cold_p_set_m (struct cold_p *p, int s, double *newm);
void cold_p_set_q (struct cold_p *p, int s, double *newq);

susc_t cold;			// TODO Nezapomnět přidat void *param do definice susc_t v trace.h

/* Nějaké příkladové implementace */

densf_t diffus_equi;
struct diffus_equi_p
{
  int s;			// Počet složek
  double t;			// Teplota
  double b;			// Referenční výška (radius)
  double gm;			// gravitační konstanta x hmotnost zdroje
  // (5.97e27 cm^3 s^-2 pro Zemi v cgs)
  double *m;			// Hmotnosti částic
  double *d_b;			// Hustoty v referenční výšce
};

void diffus_equi_p_set_s (struct diffus_equi_p *p, int news);
void diffus_equi_p_set_m (struct diffus_equi_p *p, int s, double *newm);
void diffus_equi_p_set_d_b (struct diffus_equi_p *p, int s, double *newd_b);

densf_t dens_r4;
struct dens_r4_p
{
  int s;
  double *d_1;			// Hustota 1 cm od počátku souřadnic :-)
};

void dens_r4_p_set_s (struct dens_r4_p *p, int news);
void dens_r4_p_set_d_1 (struct dens_r4_p *p, int s, double *newd_1);
void dens_r4_clean (struct dens_r4_p *p);

#include <gsl/gsl_interp.h>
densf_t dens_spline;
/*
 * Obsluhující program si musí alokovat příslušná pole funcs, accs, xas, yas,
 * sizes (v počtu s).
 */
struct dens_spline_p
{
  int s;
  gsl_interp **funcs;
  gsl_interp_accel **accs;
  double **xas;
  double **yas;
  int *sizes;
  char *file;			// Tyto dvě položky jsou zde kvůli tomu, že v přík. řádce
  int *cols;			// není jasné, zda bude první zadán soubor, nebo sloupce
};

void dens_spline_p_set_s (struct dens_spline_p *p, int news);
int dens_spline_initfun (struct dens_spline_p *p, int no,
			 const char filename[], int colx, int coly);
void dens_spline_clean (struct dens_spline_p *p);	// Nemaže samotné p

magf_t dipolz;
struct dipolz_p
{
  double moment;		// Např. 7.94e25 erg/g == 7.94e22 J/T
};

void dipolz_p_set_moment (struct dipolz_p *p, double mg_dip_moment);

#endif /* COLD_H */
