#ifndef WHAMP_H
#define WHAMP_H
#include "trace.h"
// Originál WHAMP umožňuje maximálně 6 složek
#define WHAMP_MAX_S 6

struct whamps_p;

/* parf() musí nastavit
 * v common bloku xpz_:
 * 	- aa	Nutno upozornit, že se notace liší od původního Rönnmarkova 
 * 		reportu. Namísto A(J), B(J) se používá AA(J,1), AA(J,2), resp.
 * 		aa[0][j], aa[1][j].
 * 	- dd	V Rönnmarkově reportu D
 * ve whamps_p:
 *	- t	Teplota v Kelvinech
 * 	- v_d	Rovnoběžná driftová rychlost v m/s
 * 	- n	Hustota v m^-3
 */
typedef int whamps_parf_t (rvec * r, rvec * k, struct whamps_p *param,
			   void *self_param);

susc_t whamps;
struct whamps_p
{
  // Položky nastavené cli volbami
  int s;			// Skutečný počet složek <= WHAMP_MAX_S
  int intm[WHAMP_MAX_S];	// Hmotnosti pro whamp jsou násobky protonu, 0 je elektron
  double q[WHAMP_MAX_S];	// Náboje částic
  whamps_parf_t *parf;		//Funkce, která bude krmit common bloky
  void *parf_p;
  magf_t *magf;
  void *magf_p;
  // Položky vypočtené při inicialisaci (whamps_init); pozor na jednotky
  double gyrof_wb[WHAMP_MAX_S];	// Gyrofrekvence / B_0
  double plasmaf_wn2[WHAMP_MAX_S];	// Plasmafrekvence^2 / n
  double m[WHAMP_MAX_S];	// Hmotnosti částic v kilogramech
  // Položky nastavené voláním parf
  double t[WHAMP_MAX_S];	// teplota; v_t = sqrt(2*t/m)
  double n[WHAMP_MAX_S];	// hustoty /*_U_*/
  double v_d[WHAMP_MAX_S];	// absolutní rovnoběžná driftová rychlost
  /* Položky common bloku, které má dále nastavit parf:
   * xpz_.dd[j]
   * xpz_.aa[0][j]
   * xpz_.aa[1][j]
   */
};

void whamps_p_set_s (struct whamps_p *p, int news);
void whamps_p_set_intm (struct whamps_p *p, int s, long *newm);
void whamps_p_set_q (struct whamps_p *p, int s, double *newq);
// Nastavení neměnných parametrů 
int whamps_init (struct whamps_p *p);

#include <gsl/gsl_interp.h>
// Maxwellovské isotropní rozdělení 
/* 
 * Narozdíl od dens_spline (pro studené plasma) by měly být všechny hodnoty
 * v jediném souboru
 */
whamps_parf_t whamps_parf_maxwell_spline;
struct whamps_parf_maxwell_spline_p
{
  int s;
  gsl_interp_accel *accs_temp[WHAMP_MAX_S];	// Teploty složek
  gsl_interp_accel *accs_dens[WHAMP_MAX_S];	// Hustoty složek
  gsl_interp *funcs_temp[WHAMP_MAX_S];
  gsl_interp *funcs_dens[WHAMP_MAX_S];
  double *heights;		// Pole s výškami
  double *temps[WHAMP_MAX_S];	// Teploty v dané výšce
  double *denss[WHAMP_MAX_S];	// Hustoty v dané výšce
  size_t size;			// Počet řádků
//      Jen pro inicialisaci (cli):
  int cols_temp[WHAMP_MAX_S];
  int cols_dens[WHAMP_MAX_S];
  int col_hei;
  char *file;
};

/*
 * s - skutečný počet složek
 * heicol - číslo sloupce s výškami (čísluje se od 1)
 * denscols - čísla sloupců s hustotami
 * tempcols - čísla sloupců s teplotami
 */
int whamps_parf_maxwell_spline_initfun (struct whamps_parf_maxwell_spline_p
					*p, int s, const char *filename,
					int heicol, int *denscols,
					int *tempcols);
void whamps_parf_maxwell_spline_clean (struct whamps_parf_maxwell_spline_p
				       *p);
void whamps_parf_maxwell_spline_p_set_s (struct whamps_parf_maxwell_spline_p
					 *p, int news);
void whamps_parf_maxwell_spline_p_set_cols_temp (struct
						 whamps_parf_maxwell_spline_p
						 *p, int s, long *newm);
void whamps_parf_maxwell_spline_p_set_cols_dens (struct
						 whamps_parf_maxwell_spline_p
						 *p, int s, long *newm);

double complex difu_whamp (rvec r, rvec k, double complex omega, susc_t * nic,
			   void /*struct whamps_p */ *p);
#endif
