/* TODO
 * Udělat rotování voleb (tak, aby mohly být v libovolném pořadí)
 * Sjednotit volby, které dělají to samé pro různé věci (tedy např.
 * aby volba -masses nebo -components byla platná jak pro susceptibilitu, tak pro
 * hustoty a zavolaly se postupně obě příslušné checkovací funkce
 * 
 */

#define MOC 100
#define MOCMOC 1000

#include <cli.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
struct cli_blok *
newblok (const struct cli_blok *template)
{
  struct cli_blok *neu;
  neu = malloc (sizeof (struct cli_blok));
  *neu = *template;
  neu->data = calloc (1, neu->data_sz);
  neu->optlist = malloc ((1 + neu->opt_count) * sizeof (struct cli_opt));
  // FIXME odstranit opt_count a počítat to podle OPT_EMPTY na konci
  memcpy (neu->optlist, template->optlist,
	  (1 + neu->opt_count) * sizeof (struct cli_opt));
  for (int i = 0; i < neu->opt_count; i++)
    neu->optlist[i].blok = neu;	// Nastavíme otcovský blok
  return neu;
}

/* Volat pouze na kopie, ne na šablony! */
void
deleteblok (struct cli_blok *bl)
{
  free (bl->optlist);
  free (bl->data);
  free (bl);
}

/* Registrace voleb z nového bloku. Vrací ukazatel _za_ poslední přidanou volbu. */
struct cli_opt **
opt_reg (struct cli_blok *novyblok, struct cli_opt **kam)
{
  static const struct cli_opt empty = OPT_EMPTY;
  struct cli_opt *novevolby = novyblok->optlist;
  while (memcmp (novevolby, &empty, sizeof (empty)))
    {
      *kam = novevolby;
      novevolby->blok = novyblok;
      kam++;
      novevolby++;
    }
  return kam;
}

void
print_opts (struct cli_opt **reg_opts)
{
  fprintf (stderr, "Seznam relevantních voleb:\n");
  while (*++reg_opts)
    fprintf (stderr, "-%s\n", (*reg_opts)->name);
}

void
print_choices (struct cli_opt *opt)
{
  fprintf (stderr, "Relevantní argumenty pro volbu -%s:\n", opt->name);
  const struct cli_blok **bls = opt->choices;
  while (*bls)
    {
      fprintf (stderr, "%s\n", (*bls)->name);
      ++bls;
    }
}

#include "cli_templates.h"

int
main (int argc, char **argv)
{
  if (argc < 2)
    FATAL ("Nutno zadat příkaz!\n");
  /* TODO dynamická alokace */
  struct cli_blok *reg_bloks[MOC] = { NULL };
  struct cli_opt *reg_opts[MOCMOC] = { NULL };	// Ukazatele na volby v kopiích bloku

  /* První argument vyžaduje trošičku speciální zacházení */
  struct cli_opt rootopt = { "_command_", 1, 1, NULL /*? */ ,
    (choices_t) {
		 &blok_template_density_profile,
		 &blok_template_magfield_profile,
		 &blok_template_trace1,
		 &blok_template_dispfun_view,
//                      &blok_template_whamp_dispfun_view,
		 NULL}, OPDEF
  };
  reg_opts[0] = &rootopt;
  int reg_opt_c = 1;
  struct cli_opt **reg_opts_end = reg_opts + 1;
  const struct cli_blok *curblok = pfindblok (rootopt.choices, argv[1]);
  if (!curblok)
    {
      fprintf (stderr,
	       "„%s“ není platným příkazem. Možné příkazy:\n",
	       argv[1]);
      for (int i = 0; rootopt.choices[i]->name; i++)
	fprintf (stderr, "%s\n", rootopt.choices[i]->name);
      abort ();
    }
  rootopt.count = 1;
  reg_bloks[0] = newblok (curblok);
  int reg_blok_c = 1;
  reg_opts_end = opt_reg (reg_bloks[0], reg_opts_end);

  char **argw = argv + 2;

  while (*argw)
    {
      char *arg = *argw;
      if (strncmp (arg, "-", 1))
	{
	  fprintf (stderr, "„%s“ není volba. ", arg);
	  print_opts (reg_opts);
	  abort ();
	}
      arg++;			// Počáteční pomlčka nás nezajímá
      /* nalezení příslušné volby */
      struct cli_opt *theopt = pfindpopt (reg_opts, arg);
      if (!theopt)
	{
	  fprintf (stderr, "„-%s“ není platná volba. ", arg);
	  print_opts (reg_opts);
	  abort ();
	}
      assert (theopt->choices ? theopt->has_arg : 1);	// Pokud má bloky, musí mít argument
      if (theopt->has_arg && !argw[1])
	FATAL ("Chybí argument k volbě %s!\n", argw[0]);
      /* registrace bloků */
      struct cli_blok *nbl = NULL;
      if (theopt->choices && theopt->choices[0])
	{			// registrace bloku
	  nbl = (struct cli_blok *) pfindblok (theopt->choices, argw[1]);
	  if (!nbl)
	    printchoices (theopt), abort ();
	  nbl = newblok (nbl);
	  reg_opts_end = opt_reg (nbl, reg_opts_end);
	  reg_bloks[reg_blok_c++] = nbl;
	}
      /* spouštění parsefun */
      if (theopt->parsefun)
	if (theopt->
	    parsefun (theopt->has_arg ? argw[1] : NULL, theopt->blok, nbl))
	  FATAL ("Chyba při zpracování volby %s %s\n", argw[0],
		 theopt->has_arg ? argw[1] : "");
      theopt->count++;
      argw += 1 + theopt->has_arg;
    }

  /* Kontrola, zda všechny nutné volby byly nastaveny */
  int notset = 0;
  for (struct cli_opt ** opt_i = reg_opts; opt_i < reg_opts_end; opt_i++)
    if ((*opt_i)->count < (*opt_i)->mandatory)
      fprintf (stderr, "%s-%s\n",
	       notset++ ? "" : "Některé povinné volby nebyly nastaveny:\n",
	       (*opt_i)->name);
  if (notset)
    {
      fprintf (stderr, "Nenastavené nepovinné volby:\n");
      for (struct cli_opt ** opt_i = reg_opts; opt_i < reg_opts_end; opt_i++)
	if (!(*opt_i)->count && !(*opt_i)->mandatory)
	  fprintf (stderr, "-%s\n", (*opt_i)->name);
    }
  if (notset)
    abort ();

  /* 
   * Nakonec zavoláme odzadu checkfn všech registrovaných bloků 
   * (checkfn z nultého bloku by měl provést požadovanou akci
   */
  for (int i = reg_blok_c - 1; i >= 0; i--)
    if (reg_bloks[i]->checkfn)
      if (reg_bloks[i]->checkfn (reg_bloks[i]->data))
	WARN ("Kdepak soudruzi udělali chybu?\n");

  /* Úklid, opět odzadu */
  for (int i = reg_blok_c - 1; i >= 0; i--)
    {
      if (reg_bloks[i]->data_clean)
	reg_bloks[i]->data_clean (reg_bloks[i]->data);
      deleteblok (reg_bloks[i]);
    }

  return 0;
}
