#include <cli.h>
#include <string.h>


struct cli_opt *
findopt (struct cli_opt *optptr, const char *name)
{
  const struct cli_opt empty = OPT_EMPTY;
  while (memcmp (optptr, &empty, sizeof (empty)))
    if (!strcmp (name, optptr->name))
      return optptr;
    else
      ++optptr;
  return NULL;
}


int
shortoptmatch (const char *shopt, const char *canopt)
{
  int shi = 0, cani = 0;
  while (1)
    {
      while (shopt[shi] != '\0' && shopt[shi] != '-')
	{
	  if (shopt[shi] != canopt[cani])
	    return 0;
	  shi++, cani++;
	}
      while (canopt[cani] != '\0' && canopt[cani] != '-')
	cani++;
      if (shopt[shi] == '\0')
	return 1;
      else if (canopt[cani] == '\0')
	return 0;		// Krátká delší než kanonická
      else
	shi++, cani++;
    }
}

struct cli_opt *
pfindopt (struct cli_opt *array, const char *name)
{
  /* TODO udělat to trochu sofistikovanější */
  const struct cli_opt empty = OPT_EMPTY;
  struct cli_opt *optptr = array;
  int found = 0;
  while (memcmp (optptr, &empty, sizeof (empty)))
    {
      if (shortoptmatch (name, optptr->name))
	if (!found)
	  {
	    array = optptr;
	    found = 1;
	  }
	else
	  {
	    WARN ("Volba %s je víceznačná (%s, %s).\n", name, array->name,
		  optptr->name);
	    return NULL;
	  }
      ++optptr;
    }
  if (found)
    return array;
  else
    return NULL;
}

// Totéž, ale hledá to v poli ukazatelů. Je to proto, že volby registrujeme
// pomocí ukazatelů na originály ve (zkopírovaných) blocích.
// Předchozí fce možná nakonec nebudou potřeba.
struct cli_opt *
findpopt (struct cli_opt **ptrarray, const char *name)
{
  while (ptrarray)
    {
      if (!strcmp ((*ptrarray)->name, name))
	return *ptrarray;
      ++ptrarray;
    }
  return NULL;
}

struct cli_opt *
pfindpopt (struct cli_opt **ptrarray, const char *name)
{
  /* TODO udělat to trochu sofistikovanější */
  struct cli_opt **optptr = ptrarray;
  int found = 0;
  while (NULL != *optptr)
    {
      if (shortoptmatch (name, (*optptr)->name))
	if (!found)
	  {
	    ptrarray = optptr;
	    found = 1;
	  }
	else
	  {
	    WARN ("Volba %s je víceznačná (%s, %s).\n", name,
		  (*ptrarray)->name, (*optptr)->name);
	    return NULL;
	  }
      ++optptr;
    }
  if (found)
    return *ptrarray;
  else
    return NULL;
}

const struct cli_blok *
findblok (const struct cli_blok **array, const char *name)
{
  for (int i = 0; array[i]; i++)
    if (!strcmp (name, array[i]->name))
      return array[i];
  return NULL;
}

const struct cli_blok *
pfindblok (const struct cli_blok **array, const char *name)
{
  return findblok (array, name);	// Zatím neimplementováno, je nutno napsat přesně
}

void
printchoices (const struct cli_opt *theopt)
{
  fprintf (stderr, "Platné argumenty pro volbu -%s:\n", theopt->name);
  for (const struct cli_blok ** i = theopt->choices; *i; i++)
    fprintf (stderr, "%s\n", (*i)->name);
}

/*
 * Tato funkce zjišťuje, zda je paměťová oblast o dané velikosti nulová.
 * Používá se ke zjištění, zda nějaká data již byla nastavena, či nikoliv.
 * Pokud je nulová, vrací 1, jinak 0 (pozor na odlišnost např. od memcmp).
 */
int
iszero (const void *data, size_t nbytes)
{
  const char *dc = data;
  for (size_t i = 0; i < nbytes; i++)
    if (dc[i])
      return 0;
  return 1;
}

/*********************** Obecné parsovací funkce ****************************/
#include <stdio.h>
#include <matvec.h>
#include <stdarg.h>

/* Načtení vektoru v kartézských souřadnicích ve formátu x;y;z 
 * Pokud uspěje, uloží načtenou hodnotu do o a vrátí 0, při chybě
 * vrátí -1 a nic nemění.
 *
 * Středníky mohou být problém, protože mají speciální význam
 * v shellu.
 */

int
parse_vec (const char *i, rvec * o)
{
  // LP TODO: možnost vkládání v jiném souřadném systému
  rvec result;
  double temp[3];
  if (3 != parse_doubles (i, temp, 3))
    return -1;
  else
    {
      o->val[0] = temp[0], o->val[1] = temp[1], o->val[2] = temp[2];
      return 0;
    }
}

void
FATAL (const char *fmt, ...)
{
  va_list argptr;
  va_start (argptr, fmt);
  vfprintf (stderr, fmt, argptr);
  va_end (argptr);
  abort ();
}

void
WARN (const char *fmt, ...)
{
  va_list argptr;
  va_start (argptr, fmt);
  vfprintf (stderr, fmt, argptr);
  va_end (argptr);
}

/* Zde jsou uloženy konstanty pro parse_double pro usnadnění psaní.
 * Případné delší řetězce musejí přijít před kratší.
 */
#include "consts.h"
struct
{
  const char *str;
  double val;
} pdconsts[] =
{
  {
  "qe", q_e},
  {
  "mp", m_p},
  {
  "me", m_e},
  {
  "rt", RADIUS_TERRAE},
  {
  "mt", PONDUS_TERRAE},
  {
NULL, 0}};

/* kontrola, zda na začátku řetězce není některá z pdconsts 
 * Pokud ano, vrátí index položky, pokud ne, -1
 */
int
chkpdconst (char *str)
{
  int i = 0;
  while (pdconsts[i].str)
    {
      if (!strncmp (pdconsts[i].str, str, strlen (pdconsts[i].str)))
	return i;
      i++;
    }
  return -1;
}

/* Načtení několika hodnot double oddělených středníky. Načítá maximálně
 * maxcount hodnot do pole o a vrací počet skutečně načtených hodnot
 * Pozor na speciální význam středníků!
 */
int
parse_doubles (const char *in, double o[], int maxcount)
{
  const char *pos = in;
  int count = 0;
  int pdci;
  for (count = 0; count < maxcount; count++)
    {
      if (*pos == '\0')
	break;
      char *pos_n;
      o[count] = strtod (pos, &pos_n);
      if (*pos_n == '\0')
	{
	  count++;
	  break;
	}
      else if (pos_n == pos)
	{
	  WARN ("parse_doubles(): Neočekávaný znak '%c'\n", (int) *pos_n);
	  break;
	}
      else if (*pos_n == (char) ';')
	pos = pos_n + 1;
      else if (-1 != (pdci = chkpdconst (pos_n)))
	{
	  o[count] *= pdconsts[pdci].val;
	  pos = pos_n + strlen (pdconsts[pdci].str);
	  if (*pos == (char) ';')
	    pos++;
	}
      else
	pos = pos_n;
    }
  return count;
}

/* Totéž pro long */
int
parse_longs (const char *in, long int o[], int maxcount)
{
  const char *pos = in;
  int count = 0;
  for (count = 0; count < maxcount; count++)
    {
      char *pos_n;
      o[count] = strtol (pos, &pos_n, 0);
      if (*pos_n == '\0')
	{
	  count++;		// breakem by se přeskočila inkrementace
	  break;
	}
      else if (pos_n == pos)
	{
	  WARN ("parse_longs(): Neočekávaný znak '%c'\n", (int) *pos_n);
	  break;
	}
      else if (*pos_n == (char) ';')	// To může být problém
	pos = pos_n + 1;
    }
  return count;
}
