/* 
 * Sjednotit volby, které dělají u různých věcí to samé a mají stejný formát,
 * pod stejnými názvy.
 * (A dále doplnit funkcionalitu, pochopitelně)
 * 
 */

/* 
 * V tomto souboru se nacházejí definice bloků voleb pro cli a parsovací
 * funkce k nim. Zde definované globální struktury se pak nepoužívají přímo,
 * ale kopírují se pomocí newblok() (která zároveň v kopii alokuje prostor
 * pro data atd.
 */

#define MAX_PLASMA_S 64
#define PARSE_ERROR -1
#define PARSE_SUCCESS 0

#include "cli.h"
#include "cli_templates.h"
#include "stopcond.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>


/* Užitečná makra proti zbytečnému kopírování kódu */

#define CHECKSET(var, nval)  { /* var musí být lvalue */  \
	if (!iszero(&(var), sizeof(var)))  \
	WARN("%s(): proměnná " #var " je již nastavena.",__func__); \
	var = (nval); \
}

#define CHECKDEFAULT(var, defval) {if (iszero(&(var), sizeof(var))) var = (defval); }

#define PARSE_RVEC(nazev_p, strednazvufce, var) \
	int parse_##strednazvufce##_##var (const char *opt, struct cli_blok *bl, struct cli_blok *nic) { \
		rvec neuval; \
		if (parse_vec (opt, &neuval)) return PARSE_ERROR; \
		CHECKSET(((struct nazev_p *)(bl->data))->var, neuval); \
		return PARSE_SUCCESS; \
	}

#define PARSE_DARR(nazev_p, nazevfcestred, var, varword) /* Načtení vlastností složek typu double */ \
	int parse_##nazevfcestred##_##varword (const char *optarg, struct cli_blok *bl, struct cli_blok *nic) { \
		double temp[MAX_PLASMA_S]; \
		int new_s = parse_doubles (optarg, temp, MAX_PLASMA_S); \
		nazev_p##_set_##var (bl->data, new_s, temp); \
		return PARSE_SUCCESS; \
	}

#define PARSE_COMPLEX(nazev_p, nazevfcestred, var, varword) /* Načtení double complex parametru */ \
	int parse_##nazevfcestred##_##varword (const char *optarg, struct cli_blok *bl, struct cli_blok *nic) { \
		double temp[2]; \
		int no = parse_doubles (optarg, temp, 2); \
		if(no == 1) temp[1] = 0;\
		((struct nazev_p *)(bl->data))->var = temp[0] + I*temp[1]; \
		return PARSE_SUCCESS; \
	}

#define PARSE_LIARR(nazev_p, nazevfcestred, var, varword) /* Načtení vlastností složek typu long */ \
	int parse_##nazevfcestred##_##varword (const char *optarg, struct cli_blok *bl, struct cli_blok *nic) { \
		long int temp[MAX_PLASMA_S]; \
		int new_s = parse_longs (optarg, temp, MAX_PLASMA_S); \
		nazev_p##_set_##var (bl->data, new_s, temp); \
		return PARSE_SUCCESS; \
	}

#define PARSE_DOUBLE(nazev_p, nazevfcestred, var, varword) /* Načtení jednoho double parametru */ \
	int parse_##nazevfcestred##_##varword (const char *optarg, struct cli_blok *bl, struct cli_blok *nic) { \
		char *endptr; \
		double temp = strtod(optarg, &endptr); \
		if (endptr == optarg) return PARSE_ERROR; \
		if ((int) *endptr != '\0') WARN("%s(): z argumentu „%s“ je konec „%s“ neplatný a bude ignorován.\n", __func__, optarg, endptr); \
		CHECKSET(((struct nazev_p *)(bl->data))->var, temp); \
		return PARSE_SUCCESS; \
	}

#define PARSE_INT(nazev_p, nazevfcestred, var, varword) /* Načtení jednoho int parametru */ \
	int parse_##nazevfcestred##_##varword (const char *optarg, struct cli_blok *bl, struct cli_blok *nic) { \
		char *endptr; \
		long int temp = strtol(optarg, &endptr, 0); \
		if (endptr == optarg) return PARSE_ERROR; \
		if ((int) *endptr != '\0') WARN("%s(): z argumentu „%s“ je konec „%s“ neplatný a bude ignorován.\n", __func__, optarg, endptr); \
		CHECKSET(((struct nazev_p *)(bl->data))->var, temp); \
		return PARSE_SUCCESS; \
	}

/* Funkce vytvořená tímto makrem pouze nastaví ukazatele fn a data */
#define PARSE_NBLOK(nazevrodice_p, nazevf, nazevfcestred, nazevslovy) \
	int parse_##nazevfcestred##_##nazevslovy (const char *optarg, struct cli_blok *bl, struct cli_blok *nbl) { \
		/* Nový blok by již měl mít vyplněnou fn, takže jen nastavíme ukazatele... */ \
		((struct nazevrodice_p *)(bl->data))->nazevf = nbl->fn; \
		((struct nazevrodice_p *)(bl->data))->nazevf##_p = nbl->data; \
		return PARSE_SUCCESS; \
	}

/*
 * Základní bloky: sem náležejí bloky se základními příkazy (trace,
 * density-profile, wns, ...). Příslušná volba má vždy ležet v argv[1].
 */

/* ========================= Blok magfield_profile ======================= */

int
commit_magfield_profile (void *data)
{
  struct magfield_profile_p *p = data;
  // Výchozí nastavení je 100 bodů výpisu:
  rvec stepvec = rvec_sub (&(p->to), &(p->from));
  double dist = rvec_norma (&stepvec);
  if (p->stepcount)
    {
      if (p->step)
	fprintf (stderr,
		 "%s: Nastavena délka kroku i počet. Počet má přednost.\n",
		 __func__);
      p->step = dist / p->stepcount;
    }
  else if (p->step)
    p->stepcount = dist / p->step;
  else
    {
      p->stepcount = 99;
      p->step = dist / 99;
    }
  stepvec = rvec_normuj (&stepvec);
  stepvec = rvec_mul_con (p->step, &stepvec);

  for (int i = 0; i < p->stepcount + 1; i++)
    {
      rvec kde = rvec_mul_con (i, &stepvec);
      kde = rvec_add (&(p->from), &kde);
      rvec b = p->magf (&kde, p->magf_p);
      printf ("%le ", p->step * i);
      for (int j = 0; j < 3; j++)
	printf ("%le ", kde.val[j]);
      for (int j = 0; j < 3; j++)
	printf ("%le ", b.val[j]);
      putchar ('\n');
    }
  return 0;
}

PARSE_RVEC (magfield_profile_p, magfield_profile, from);
PARSE_RVEC (magfield_profile_p, magfield_profile, to);
PARSE_DOUBLE (magfield_profile_p, magfield_profile, step, step);
PARSE_INT (magfield_profile_p, magfield_profile, stepcount, stepcount);
PARSE_NBLOK (magfield_profile_p, magf, magfield_profile, magfield);

const struct cli_blok blok_template_magfield_profile = {
  "magfield-profile",
  (optlist_t) {			// optlist
	       {"magfield", 1, 1, parse_magfield_profile_magfield, (choices_t) {
										&blok_template_magf_dipolz,
										&blok_template_magf_igrf,
										NULL},
		OPDEF},
	       {"from", 1, 1, parse_magfield_profile_from, NULL, OPDEF},
	       {"to", 1, 1, parse_magfield_profile_to, NULL, OPDEF},
	       {"step-length", 0, 1, parse_magfield_profile_step, NULL,
		OPDEF},
	       {"step-count", 0, 1, parse_magfield_profile_stepcount, NULL,
		OPDEF},
	       OPT_EMPTY}, 5,
  NULL,				// Vykonává to commit_magfield_profile
  sizeof (struct magfield_profile_p),
  commit_magfield_profile,
  NULL,				// Destruktor není třeba
  NULL
};

/* ========================= Blok density_profile ======================== */
/*
 * DŮLEŽITÉ: První položkou datové struktury pro hustotu musí být _vždy_ počet
 * komponent (typu int).
 */
int
commit_density_profile (void *data)
{
  struct density_profile_p *p = data;
  // Výchozí nastavení je 100 (99?) bodů výpisu:
  CHECKDEFAULT (p->step, rvec_dist (&(p->from), &(p->to)) / 100);
  int s = *((int *) (p->densf_p));	// s musí být PRVNÍ !!!

  double d[s];

  rvec stepvec = rvec_sub (&(p->to), &(p->from));
  stepvec = rvec_normuj (&stepvec);
  stepvec = rvec_mul_con (p->step, &stepvec);
  int count = rvec_dist (&(p->from), &(p->to)) / p->step;

  for (int i = 0; i < count; i++)
    {
      rvec kde = rvec_mul_con (i, &stepvec);
      kde = rvec_add (&(p->from), &kde);
      p->densf (d, &kde, p->densf_p);
      printf ("%le ", p->step * i);
      for (int j = 0; j < s - 1; j++)
	printf ("%le ", d[j]);
      printf ("%le\n", d[s - 1]);
    }
  return 0;
}

PARSE_RVEC (density_profile_p, density_profile, from);
PARSE_RVEC (density_profile_p, density_profile, to);
PARSE_DOUBLE (density_profile_p, density_profile, step, step);
PARSE_NBLOK (density_profile_p, densf, density_profile, density);

const struct cli_blok blok_template_dens_spline;

const struct cli_blok blok_template_density_profile = {
  "density-profile",
  (optlist_t) {			// optlist
	       {"density", 1, 1, parse_density_profile_density, (choices_t) {
									     &blok_template_dens_spline,
									     &blok_template_dens_r4,
									     &blok_template_dens_difeq,
									     &blok_template_dens_iri,
									     NULL},
		OPDEF},
	       {"from", 1, 1, parse_density_profile_from, NULL, OPDEF},
	       {"to", 1, 1, parse_density_profile_to, NULL, OPDEF},
	       {"step", 0, 1, parse_density_profile_step, NULL, OPDEF},
	       OPT_EMPTY}, 4,
  NULL,				// Vykonává to commit_density_profile
  sizeof (struct density_profile_p),
  commit_density_profile,
  NULL,				// Destruktor není třeba
  NULL
};

/********************************************
 * Bloky se stop-podmínkami pro ray tracing *
 ********************************************/
#include "stopcond.h"

/* ========================= Blok stop_equator =========================== */

const struct cli_blok blok_template_stop_equator = {
  "equator",
  (optlist_t) {
	       OPT_EMPTY}, 0,
  (void (*)()) stop_equator,
  sizeof (struct stop_equator_p),
  NULL,				// Není třeba, data se callocují, takže je to 0, jak má být
  NULL,
  NULL
};

/* ========================= Blok stop_counter =========================== */

int
parse_stop_counter_step_count (const char *optarg, struct cli_blok *bl,
			       struct cli_blok *nic)
{
  CHECKSET (((struct stop_counter_p *) (bl->data))->steps_left,
	    atoi (optarg));
  if (((struct stop_counter_p *) (bl->data))->steps_left <= 0L)
    {
      WARN ("%s(): počet kroků musí býti kladný.\n");
      return PARSE_ERROR;
    }
  else
    return PARSE_SUCCESS;
}

const struct cli_blok blok_template_stop_counter = {
  "counter",
  (optlist_t) {
	       {"step-count", 1, 1, parse_stop_counter_step_count, NULL,
		OPDEF},
	       OPT_EMPTY}, 1,
  (void (*)()) stop_counter,
  sizeof (struct stop_counter_p),
  NULL,				// Vše je povinné, dál není co řešit
  NULL,				// Destruktor není třeba
  NULL
};

/* ====================== Blok stop_radial =============================== */

PARSE_DOUBLE (stop_radial_p, stop_radial, outer_boundary_radius,
	      outer_boundary);
PARSE_DOUBLE (stop_radial_p, stop_radial, inner_boundary_radius,
	      inner_boundary);

const struct cli_blok blok_template_stop_radial = {
  "radial",
  (optlist_t) {
	       {"inner-boundary", 1, 1, parse_stop_radial_inner_boundary,
		NULL, OPDEF},
	       {"outer-boundary", 1, 1, parse_stop_radial_outer_boundary,
		NULL, OPDEF},
	       OPT_EMPTY}, 2,
  (void (*)()) stop_radial,
  sizeof (struct stop_radial_p),
  NULL,				// Vše je povinné, dál není co řešit
  NULL,				// Destruktor není třeba
  NULL
};

/* ========================= Blok dispfun_view =========================== */
// Vypíše pro pevné r, ω a směr K dispersní funkci G(r, |k|K, ω) v závislosti
// na |k|.
PARSE_COMPLEX (dispfun_view_p, dispfun_view, frequency, frequency);
PARSE_DOUBLE (dispfun_view_p, dispfun_view, from, from);
PARSE_DOUBLE (dispfun_view_p, dispfun_view, to, to);
PARSE_DOUBLE (dispfun_view_p, dispfun_view, step, step);
PARSE_RVEC (dispfun_view_p, dispfun_view, position);
PARSE_RVEC (dispfun_view_p, dispfun_view, direction);
PARSE_NBLOK (dispfun_view_p, susc, dispfun_view, susceptibility);

int
commit_dispfun_view (void *data)
{
  struct dispfun_view_p *p = data;
  p->direction = rvec_normuj (&(p->direction));
  for (double k = p->from; k <= p->to; k += p->step)
    {
      complex double gg =
	g (p->position, rvec_mul_con (k, &(p->direction)), p->frequency,
	   p->susc, p->susc_p);
      printf ("%e %e %e\n", k, creal (gg), cimag (gg));
    }
  return 0;
}

const struct cli_blok blok_template_dispfun_view = {
  "dispfun-view",
  (optlist_t) {
	       {"position", 1, 1, parse_dispfun_view_position, NULL, OPDEF},
	       {"susceptibility", 1, 1, parse_dispfun_view_susceptibility,
		(choices_t) {&blok_template_susc_cold,
			     &blok_template_susc_whamps, NULL}, OPDEF},
	       {"frequency", 1, 1, parse_dispfun_view_frequency, NULL, OPDEF},
	       {"direction", 1, 1, parse_dispfun_view_direction, NULL, OPDEF},
	       {"from", 1, 1, parse_dispfun_view_from, NULL, OPDEF},
	       {"to", 1, 1, parse_dispfun_view_to, NULL, OPDEF},
	       {"step", 1, 1, parse_dispfun_view_step, NULL, OPDEF},
	       NULL}, 7,
  NULL,				// Vykonává to commit_dispfun_view
  sizeof (struct dispfun_view_p),
  commit_dispfun_view,
  NULL,
  NULL
};

/* ========================= Blok trace ================================== */
//suffix trace1 je uveden, aby se to nepletlo s nízkoúrovňovým trace a trace_p

PARSE_DOUBLE (trace1_p, trace, step, step);
PARSE_DOUBLE (trace1_p, trace, maxt, maxt);
PARSE_COMPLEX (trace1_p, trace, frequency, frequency);
PARSE_RVEC (trace1_p, trace, from);
PARSE_RVEC (trace1_p, trace, direction);
/* Zde kompilátor zahlásí varování kvůli jinému typu funkce, ale to ničemu nevadí */
PARSE_NBLOK (trace1_p, susc, trace, susceptibility);
PARSE_NBLOK (trace1_p, stop_condition, trace, stop_condition);	// TODO blok_template_stop_condition

int
parse_trace_mode (const char *optarg, struct cli_blok *bl,
		  struct cli_blok *nic)
{
  if (!strncmp (optarg, "inner", sizeof ("inner")) || atoi (optarg) == 1)
    {
      ((struct trace1_p *) (bl->data))->mod = 1;
      return PARSE_SUCCESS;
    }
  else if (!strncmp (optarg, "outer", sizeof ("outer"))
	   || atoi (optarg) == -1)
    {
      ((struct trace1_p *) (bl->data))->mod = -1;
      return PARSE_SUCCESS;
    }
  else
    return PARSE_ERROR;
}

int
commit_trace1 (void *data)
{
  struct trace1_p *p = data;
  if (!p->step)
    p->step = 1e-3;		// milisekunda
  if (!p->maxt)
    p->maxt = HUGE_VAL;
  if (!p->mod)
    {
      WARN
	("%s(): Není nastaveno, které řešení dispersní rovnice použít."
	 "Používám vnitřní řešení (nejmenší |k|).\n", __func__);
      p->mod = 1;
    }
  trace (&(p->from), &(p->direction), p->frequency, p->stop_condition,
	 p->stop_condition_p, p->maxt, p->step, p->susc, p->susc_p, p->mod);
  return 0;
}

const struct cli_blok blok_template_trace1 = {
  "trace",
  (optlist_t) {
	       {"susceptibility", 1, 1, parse_trace_susceptibility,
		(choices_t) {&blok_template_susc_cold,
			     &blok_template_susc_whamps, NULL}, OPDEF},
	       {"from", 1, 1, parse_trace_from, NULL, OPDEF},
	       {"direction", 1, 1, parse_trace_direction, NULL, OPDEF},
	       {"stop-condition", 1, 1, parse_trace_stop_condition,
		(choices_t) {&blok_template_stop_equator,
			     &blok_template_stop_counter,
			     &blok_template_stop_radial, NULL}, OPDEF},
	       {"frequency", 1, 1, parse_trace_frequency, NULL, OPDEF},
	       {"step-time", 0, 1, parse_trace_step, NULL, OPDEF},
	       {"maximal-time", 0, 1, parse_trace_maxt, NULL, OPDEF},
	       {"mode", 0, 1, parse_trace_mode, NULL, OPDEF},
	       OPT_EMPTY}, 8,
  NULL,				// Vykonává to commit_trace1
  sizeof (struct trace1_p),
  commit_trace1,
  NULL,				// Destruktor není třeba
  NULL
};

#ifndef NOT_USE_SUSCEPTIBILITY_COLD
/*
 * Bloky obsluhující funkce se základním studeným plasmatem (cold.h)
 */

#define PARSE_COMPONENTS(nazev_p, strednazvufce)\
	int parse_##strednazvufce##_components (const char *optarg, struct cli_blok *bl, struct cli_blok *nic) { \
		long new_s; \
		if (1 != parse_longs (optarg, &new_s, 1)) return -1; \
		/* TODO zde kontrola konsistence */ \
		nazev_p##_set_s (bl->data, new_s); \
		return PARSE_SUCCESS; \
	}

#include <cold.h>

/* ========================= Blok dens_spline ============================ */

/* Soubor s interpolovanými body */
int
parse_dens_spline_file (const char *optarg, struct cli_blok *bl,
			struct cli_blok *nic)
{
  size_t len = strlen (optarg);
  ((struct dens_spline_p *) (bl->data))->file =
    realloc (((struct dens_spline_p *) (bl->data))->file,
	     len + sizeof (char));
  strcpy (((struct dens_spline_p *) (bl->data))->file, optarg);
  // TODO? kontrola vícenásobného nastavení...
  return 0;
}

/* Čísla sloupců oddělená středníky. První je výška, ostatní hustoty. Od nuly. */
int
parse_dens_spline_columns (const char *optarg, struct cli_blok *bl,
			   struct cli_blok *nic)
{
  long temp[MAX_PLASMA_S];
  int new_s1 = parse_longs (optarg, temp, MAX_PLASMA_S);
  ((struct dens_spline_p *) (bl->data))->cols =
    realloc (((struct dens_spline_p *) (bl->data))->cols,
	     new_s1 * sizeof (int));
  for (int i = 0; i < new_s1; i++)
    ((struct dens_spline_p *) (bl->data))->cols[i] = temp[i];
  dens_spline_p_set_s (bl->data, new_s1 - 1);
  return 0;
}

PARSE_COMPONENTS (dens_spline_p, dens_spline);

/* 
 * Výchozí chování: není-li počet složek nijak nastaven, berou se sloupce 0,1 
 * Nejsou-li nastaveny sloupce, berou se po řadě 0, 1, ..., s + 1.
 */
int
check_dens_spline (void *data)
{
  struct dens_spline_p *p = data;
  if (!p->s)
    dens_spline_p_set_s (p, 1);
  if (!p->cols)
    {
      p->cols = calloc (p->s + 1, sizeof (int));
      for (int i = 0; i < p->s + 1; i++)
	p->cols[i] = i;
    }
  p->funcs = calloc (p->s + 1, sizeof (gsl_interp *));
  p->accs = calloc (p->s + 1, sizeof (gsl_interp_accel *));
  p->xas = calloc (p->s + 1, sizeof (double *));
  p->yas = calloc (p->s + 1, sizeof (double *));
  p->sizes = calloc (p->s + 1, sizeof (int));
  for (int i = 0; i < p->s; i++)
    // Pozor, tato funkce nekontroluje validitu vstupního souboru!
    dens_spline_initfun (p, i, p->file, p->cols[0], p->cols[i + 1]);
  return 0;
}

void
clean_dens_spline (void *data)
{
  struct dens_spline_p *p = data;
  dens_spline_clean (p);
  free (p->funcs);
  free (p->accs);
  free (p->xas);
  free (p->yas);
  free (p->sizes);
}

const struct cli_blok blok_template_dens_spline = {
  "spline",
  (optlist_t) {			// optlist
	       {"spline-file", 1, 1, parse_dens_spline_file, NULL, OPDEF},
	       {"spline-columns", 0, 1, parse_dens_spline_columns, NULL,
		OPDEF},
	       {"spline-components", 0, 1, parse_dens_spline_components, NULL,
		OPDEF},
	       OPT_EMPTY}, 3,
  dens_spline,
  sizeof (struct dens_spline_p),
  check_dens_spline,
  clean_dens_spline,
  NULL
};

/* ========================= Blok dens_r4 ================================ */

PARSE_COMPONENTS (dens_r4_p, dens_r4);
PARSE_DARR (dens_r4_p, dens_r4, d_1, base_density);

const struct cli_blok blok_template_dens_r4 = {
  "r4",
  (optlist_t) {			// optlist
	       {"r4-components", 0, 1, parse_dens_r4_components, NULL, OPDEF},
	       {"r4-density-base", 1, 1, parse_dens_r4_base_density, NULL,
		OPDEF},
	       OPT_EMPTY}, 2,
  dens_r4,
  sizeof (struct dens_r4_p),
  NULL /* check_dens_r4 - není co ověřovat */ ,
  (void (*)(void *)) dens_r4_clean,
  NULL
};

/* ========================= Blok dens_difeq ============================= */

PARSE_COMPONENTS (diffus_equi_p, dens_difeq);
PARSE_DARR (diffus_equi_p, dens_difeq, d_b, base_density);
PARSE_DARR (diffus_equi_p, dens_difeq, m, masses);
PARSE_DOUBLE (diffus_equi_p, dens_difeq, b, base);
PARSE_DOUBLE (diffus_equi_p, dens_difeq, t, temperature);
/* Trošku prasárna, gravitační konstantou se to vynásobí až při check_dens_difeq: */
PARSE_DOUBLE (diffus_equi_p, dens_difeq, gm, planet_mass);

int
check_dens_difeq (void *data)
{
  struct diffus_equi_p *p = data;
  if (!p->gm)
    p->gm = 5.9736e24;		// Hmotnost Země v kg
  p->gm *= GSL_CONST_MKSA_GRAVITATIONAL_CONSTANT;
  /* momentálně mě nic dalšího nenapadá, kromě nějakých záporných hodnot */
  return 0;
}

void
clean_dens_difeq (void *data)
{
  struct diffus_equi_p *p = data;
  free (p->m);
  free (p->d_b);
}

const struct cli_blok blok_template_dens_difeq = {
  "diffusive-equilibrium",
  (optlist_t) {
	       {"difeq-components", 0, 1, parse_dens_difeq_components, NULL,
		OPDEF},
	       {"difeq-planet-mass", 0, 1, parse_dens_difeq_planet_mass, NULL,
		OPDEF},
	       {"difeq-density-base", 1, 1, parse_dens_difeq_base_density,
		NULL, OPDEF},
	       {"difeq-base-level", 1, 1, parse_dens_difeq_base, NULL, OPDEF},
	       {"difeq-temperature", 1, 1, parse_dens_difeq_temperature, NULL,
		OPDEF},
	       {"difeq-masses", 1, 1, parse_dens_difeq_masses, NULL, OPDEF},
	       OPT_EMPTY}, 6,
  (void (*)()) diffus_equi,
  sizeof (struct diffus_equi_p),
  check_dens_difeq,
  clean_dens_difeq,
  NULL
};

/* ========================= Blok dens_iri =============================== */
#include "iridens.h"

PARSE_COMPONENTS (iri_p, dens_iri);

int
parse_dens_iri_ion_names (const char *optarg, struct cli_blok *bl,
			  struct cli_blok *nic)
{
  // Mimochodem, nekontroluje to jednoznačnost
  char buf[8];
  int pos = 0;
  for (int typecount = 0; typecount < IRIDENS_MAX_S; typecount++)
    {
      int bufi = 0;
      while (optarg[pos] != (char) '\0' && optarg[pos] != (char) ';')
	{
	  buf[bufi] = optarg[pos];
	  bufi++, pos++;
	  if (bufi > 8)
	    {
	      WARN ("%s: Neznámý typ částice: %.8s\n", __func__, buf);
	      return PARSE_ERROR;
	    }
	}
      buf[bufi] = (char) '\0';
      for (int i = 0; i < IRIDENS_MAX_S /*? */ ; i++)
	if (strcmp (buf, iri_ion_names[i]))
	  {
	    ((struct iri_p *) (bl->data))->mapping[typecount] = i;
	    goto parse_dens_iri_ion_names_shoda;
	  }
      WARN ("%s: Neznámý typ částice: %s\n", __func__, buf);
      return PARSE_ERROR;
    parse_dens_iri_ion_names_shoda:
      if (optarg[pos] == (char) ';')
	pos++;
      else if (optarg[pos] == (char) '\0')
	{
	  iri_p_set_s ((struct iri_p *) (bl->data), typecount + 1);
	  return PARSE_SUCCESS;
	}
    }
  WARN ("%s: IRI rozeznává pouze %s druhů částic, zbytek ignoruji.\n",
	__func__, IRIDENS_MAX_S);
  iri_p_set_s ((struct iri_p *) (bl->data), IRIDENS_MAX_S);
  return PARSE_SUCCESS;
}

#define CHECKLIM(podm)  \
	if(!(podm)) { \
		WARN("%s: Zadané datum %s porušuje podmínku pro rozsah " #podm "\n", __func__, optarg); \
		return PARSE_ERROR; \
	}

int
parse_dens_iri_date (const char *optarg, struct cli_blok *bl,
		     struct cli_blok *nic)
{
  struct iri_p *p = bl->data;
  int y, mo, d, h, m = 0, s = 0, daysum = 0;
  int matched = sscanf (optarg, "%d-%d-%d %d:%d:%d", &y, &mo, &d, &h, &m, &s);
  if (matched < 4)
    {
      WARN
	("%s: Nesprávný formát času \"%s\". Má být \"RRRR-MM-DD hh:mm:ss\".\n",
	 __func__, optarg);
      return PARSE_ERROR;
    }
  CHECKLIM (mo > 0);
  CHECKLIM (mo <= 12);
  CHECKLIM (h >= 0);
  CHECKLIM (h < 24);
  CHECKLIM (m >= 0);
  CHECKLIM (m < 60);
  CHECKLIM (s >= 0);
  CHECKLIM (s < 60);

  int mdaycount[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
  if (0 == y % 4)
    mdaycount[1] = 29;
  if (0 == y % 100)
    mdaycount[1] = 28;

  p->year = y;
  CHECKLIM (d > 0);
  CHECKLIM (d <= mdaycount[mo - 1]);
  for (int i = 1; i < mo; i++)
    daysum += mdaycount[i - 1];
  daysum += d;
  p->day = daysum;
  p->hour = h + m / 60. + s / 3600.;
  return PARSE_SUCCESS;
}

int
parse_dens_iri_path (const char *optarg, struct cli_blok *bl,
		     struct cli_blok *nic)
{
  iri_data_path = optarg;
  return 0;
}

int
check_dens_iri (void *data)
{
  struct iri_p *p = data;
  iriwrap_iyyyy = p->year;
  iriwrap_mmdd = -p->day;
  iriwrap_dhour = p->hour + 25;
  return 0;
}

const struct cli_blok blok_template_dens_iri = {
  "iri",
  (optlist_t) {
	       {"iri-components", 0, 1, parse_dens_iri_components, NULL, OPDEF},	// Asi úplně zbytečné, jako všude jinde
	       {"iri-ion-types", 1, 1, parse_dens_iri_ion_names, NULL, OPDEF},
	       {"iri-date", 1, 1, parse_dens_iri_date, NULL, OPDEF},
	       {"iri-path", 0, 1, parse_dens_iri_path, NULL, OPDEF},
	       OPT_EMPTY}, 4,
  (void (*)()) iri,
  sizeof (struct iri_p),
  check_dens_iri,
  NULL,
  NULL
};

/* ========================= Blok magf_dipolz ============================ */

int
parse_magf_dipolz_moment (const char *optarg, struct cli_blok *bl,
			  struct cli_blok *nic)
{
  double temp;
  if (1 != parse_doubles (optarg, &temp, 1))
    {
      WARN ("%s(): očekávám jedno desetinné číslo (zadáno „%s“).",
	    __func__, optarg);
      return -1;
    }
  dipolz_p_set_moment (bl->data, temp);
  return 0;
}

int
check_magf_dipolz (void *data)
{
  struct dipolz_p *p = data;
  if (!p->moment)
    dipolz_p_set_moment (p, 0);	// Nastaví výchozí mag. dip. moment
  return 0;
}

const struct cli_blok blok_template_magf_dipolz = {
  "dipolz",
  (optlist_t) {			// optlist
	       {"moment", 0, 1, parse_magf_dipolz_moment, NULL, OPDEF},
	       OPT_EMPTY}, 1,
  (void (*)()) dipolz,
  sizeof (struct dipolz_p),
  check_magf_dipolz,
  NULL,				// Destruktor zde není potřeba
  NULL
};

/* ========================= Blog magf_igrf ============================== */
#include "igrf.h"

int
check_magf_igrf (void *data)
{
  struct igrf_p *p = data;
  igrf_checkinit (p);
  return 0;
}

int
parse_magf_igrf_path (const char *optarg, struct cli_blok *bl,
		      struct cli_blok *nic)
{
  igrf_data_path = optarg;
  return 0;
}


int
parse_magf_igrf_date (const char *optarg, struct cli_blok *bl,
		      struct cli_blok *nic)
{
  struct igrf_p *p = bl->data;
  int y, mo = 1, d = 1, h = 0, m = 0, s = 0, daysum = 0;
  int matched = sscanf (optarg, "%d-%d-%d %d:%d:%d", &y, &mo, &d, &h, &m, &s);
  if (matched < 2)
    {				// Měl by být alespoň rok a měsíc
      WARN
	("%s: Nesprávný formát času \"%s\". Má být \"RRRR-MM[-DD[ hh[:mm[:ss]]]]\".\n",
	 __func__, optarg);
      return PARSE_ERROR;
    }
  CHECKLIM (mo > 0);
  CHECKLIM (mo <= 12);
  CHECKLIM (h >= 0);
  CHECKLIM (h < 24);
  CHECKLIM (m >= 0);
  CHECKLIM (m < 60);
  CHECKLIM (s >= 0);
  CHECKLIM (s < 60);

  real yeardays = 365;
  int mdaycount[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
  if (0 == y % 4)
    mdaycount[1] = 29, yeardays = 366;
  if (0 == y % 100)
    mdaycount[1] = 28, yeardays = 365;

  CHECKLIM (d > 0);
  CHECKLIM (d <= mdaycount[mo - 1]);
  for (int i = 1; i < mo; i++)
    daysum += mdaycount[i - 1];
  daysum += d;
  p->year = (real) y + (mo - 1) / 12. + (daysum - 1) / yeardays
    + h / (24 * yeardays) + m / (60 * 24 * yeardays) +
    s / (60 * 60 * 24 * yeardays);
  return PARSE_SUCCESS;
}

const struct cli_blok blok_template_magf_igrf = {
  "igrf",
  (optlist_t) {
	       {"igrf-date", 1, 1, parse_magf_igrf_date, NULL, OPDEF},
	       {"igrf-path", 0, 1, parse_magf_igrf_path, NULL, OPDEF},
	       OPT_EMPTY}, 2,
  (void (*)()) igrf,
  sizeof (struct igrf_p),
  check_magf_igrf,
  NULL,
  NULL
};

/* ========================= Blok susc_cold ============================== */

PARSE_COMPONENTS (cold_p, susc_cold);
PARSE_DARR (cold_p, susc_cold, m, masses);
PARSE_DARR (cold_p, susc_cold, q, charges);
PARSE_NBLOK (cold_p, densf, susc_cold, density);
/* Zde kompilátor zahlásí varování kvůli jinému typu funkce, ale to ničemu nevadí */
PARSE_NBLOK (cold_p, magf, susc_cold, magfield);

const struct cli_blok blok_template_susc_cold = {
  "cold",
  (optlist_t) {			//optlist:
	       {"components", 0, 1, parse_susc_cold_components, NULL, OPDEF},
	       {"masses", 1, 1, parse_susc_cold_masses, NULL, OPDEF},
	       {"charges", 1, 1, parse_susc_cold_charges, NULL, OPDEF},
	       {"density", 1, 1, parse_susc_cold_density,
		(choices_t) {&blok_template_dens_spline,
			     &blok_template_dens_r4,
			     &blok_template_dens_difeq, NULL}, OPDEF},
	       {"magfield", 1, 1, parse_susc_cold_magfield,
		(choices_t) {&blok_template_magf_dipolz,
			     &blok_template_magf_igrf, NULL}, OPDEF},
	       OPT_EMPTY},
  5,
  (void (*)()) cold,
  sizeof (struct cold_p),
  NULL,				//není co inicialisovat
  NULL,				//„destruktor“ není třeba
  NULL
};

#endif // USE_SUSCEPTIBILITY_COLD

/****************************** Bloky pro WHAMP ****************************/
#include "whamp.h"
/* ========================= Blok whamps_parf_maxwell_spline ============= */

PARSE_COMPONENTS (whamps_parf_maxwell_spline_p, whamps_parf_maxwell_spline);
int
check_whamps_parf_maxwell_spline (void *data)
{
  struct whamps_parf_maxwell_spline_p *p = data;
  return whamps_parf_maxwell_spline_initfun (p, p->s, p->file, p->col_hei,
					     p->cols_dens, p->cols_temp);
}

/* Soubor s interpolovanými body */
int
parse_whamps_parf_maxwell_spline_filename (const char *optarg,
					   struct cli_blok *bl,
					   struct cli_blok *nic)
{
  size_t len = strlen (optarg);
  ((struct whamps_parf_maxwell_spline_p *) (bl->data))->file =
    realloc (((struct whamps_parf_maxwell_spline_p *) (bl->data))->file,
	     len + sizeof (char));
  strcpy (((struct whamps_parf_maxwell_spline_p *) (bl->data))->file, optarg);
  // TODO? kontrola vícenásobného nastavení...
  return 0;
}

/* Čísla sloupců oddělená středníky. Od jedné? */
PARSE_LIARR (whamps_parf_maxwell_spline_p, whamps_parf_maxwell_spline,
	     cols_dens, density_columns);
PARSE_LIARR (whamps_parf_maxwell_spline_p, whamps_parf_maxwell_spline,
	     cols_temp, temperature_columns);
PARSE_INT (whamps_parf_maxwell_spline_p, whamps_parf_maxwell_spline, col_hei,
	   radius_column);

void
clean_whamps_parf_maxwell_spline (void *data)
{
  struct whamps_parf_maxwell_spline_p *p = data;
  free (p->file);
  whamps_parf_maxwell_spline_clean (p);
}

const struct cli_blok blok_template_whamps_parf_maxwell_spline = {
  "maxwell-spline",
  (optlist_t) {
	       {"components", 0, 1,
		parse_whamps_parf_maxwell_spline_components, NULL, OPDEF},
	       {"spline-filename", 1, 1,
		parse_whamps_parf_maxwell_spline_filename, NULL, OPDEF},
	       {"temperature-columns", 1, 1,
		parse_whamps_parf_maxwell_spline_temperature_columns, NULL,
		OPDEF},
	       {"density-columns", 1, 1,
		parse_whamps_parf_maxwell_spline_density_columns, NULL,
		OPDEF},
	       {"radius-column", 1, 1,
		parse_whamps_parf_maxwell_spline_radius_column, NULL, OPDEF},
	       OPT_EMPTY},
  5,
  (void (*)()) whamps_parf_maxwell_spline,
  sizeof (struct whamps_parf_maxwell_spline_p),
  check_whamps_parf_maxwell_spline,
  clean_whamps_parf_maxwell_spline,
  NULL
};

/* ========================= Blok susc_whamps ============================ */
PARSE_COMPONENTS (whamps_p, susc_whamps);
PARSE_DARR (whamps_p, susc_whamps, q, charges);
PARSE_LIARR (whamps_p, susc_whamps, intm, masses);
PARSE_NBLOK (whamps_p, parf, susc_whamps, velocity_distribution);
PARSE_NBLOK (whamps_p, magf, susc_whamps, magfield);

int
check_susc_whamps (void *data)
{
  struct whamps_p *p = data;
  g = difu_whamp;		// Mění se nastavení dispersní funkce!
  return whamps_init (p);
}

const struct cli_blok blok_template_susc_whamps = {
  "whamp",
  (optlist_t) {
	       {"components", 0, 1, parse_susc_whamps_components, NULL,
		OPDEF},
	       {"masses", 1, 1, parse_susc_whamps_masses, NULL, OPDEF},
	       {"charges", 1, 1, parse_susc_whamps_charges, NULL, OPDEF},
	       {"velocity-distribution", 1, 1,
		parse_susc_whamps_velocity_distribution,
		(choices_t) {&blok_template_whamps_parf_maxwell_spline, NULL},
		OPDEF},
	       {"magfield", 1, 1, parse_susc_whamps_magfield,
		(choices_t) {&blok_template_magf_dipolz,
			     &blok_template_magf_igrf, NULL}, OPDEF},
	       OPT_EMPTY},
  5,
  (void (*)()) whamps,
  sizeof (struct whamps_p),
  check_susc_whamps,
  NULL,				// Není co uvolňovat
  NULL
};
