
#define dbg stderr

#include "cold.h"
#include <gsl/gsl_const_cgs.h>
#include <math.h>
#include <gsl/gsl_errno.h>
#include <string.h>
#include <stdio.h>		//pro ladění

#include "consts.h"

#define MACRO_ENABLE_najdisloupec
#define MACRO_ENABLE_dojedradek
#include "commacros.h"

/* Susceptibilita studeného plasmatu */
mat
cold (rvec * r, rvec * k, double omega, void *susc_p)
{
  struct cold_p *p = susc_p;
  int s = p->s;
  double *q = p->q;		/* Náboj částice */
  double *m = p->m;		/* Hmotnost částice (přibližně zatím) */
  double n[s];			/* Hustota */
  p->densf (n, r, p->densf_p);

  double b_0;			// Magnetická indukce v nultém řádu
  rvec b, b_smer;
  b = p->magf (r, p->magf_p);
  b_smer = rvec_normuj (&b);	// TODO možná nadbytečné.
  b_0 = rvec_norma (&b);

  /* ω_p^2 kvadrát plasmové freq, Ω cyklotronová freq. */
  double omega2_p[s], Omega[s];
  for (int i = 0; i < s; i++)
    {
      omega2_p[i] = n[i] * SQ (q[i]) / (m[i] * epsilon_0);
      Omega[i] = q[i] * b_0 / m[i];
    }

  double Rt = 0, Lt = 0, Pt = 0;
  for (int i = 0; i < s; i++)
    {
      Rt += -omega2_p[i] / (omega * (omega + Omega[i]));
      Lt += -omega2_p[i] / (omega * (omega - Omega[i]));
      Pt += -omega2_p[i] / SQ (omega);
#ifdef DEBUG_SUSC
      fprintf (dbg, "cold: R-1 = %e, L-1 = %e, P-1 = %e, Ω[%d] = %e\n", Rt,
	       Lt, Pt, i, Omega[i]);
#endif
    }
  double St = (Rt + Lt) / 2, Dt = (Rt - Lt) / 2;
  mat sus = { {{St, -I * Dt, 0}, {I * Dt, St, 0}, {0, 0, Pt}} };
  /* Převést do správné base */
  mat tf = ortho_xz (&b_smer, k);
  sus = mat_mul (&sus, &tf);
  /* Tohle by šlo optimalisovat nějakou separátní funkcí */
  tf = mat_tr (&tf);
  sus = mat_mul (&tf, &sus);
  return sus;
}

void
dipolz_p_set_moment (struct dipolz_p *p, double moment)
{
  if (p->moment)
    fprintf (stderr, "%s(): "
	     "Varování: změna magnet. momentu (z %lf na %lf).\n", __func__,
	     p->moment, moment);
  if (!moment)
    {
      moment = 7.94e22 /* A m^2 */ ;	//7.94e25 /* erg/g */; // „Tabulkový“ mg. moment Země
      fprintf (stderr, "%s(): "
	       "Varování: zadán nulový magnetický moment, nastavuji na %le.\n",
	       __func__, moment);
    }
  p->moment = moment;
}

#define SET_S(nazev_p) void 						\
nazev_p##_set_s (struct nazev_p *p, int news) 				\
{ 									\
  if (news <= 0) 							\
    { 									\
      fprintf (stderr, "%s(): Počet složek má býti kladný!\n",__func__);\
      abort ();								\
    }									\
  else if (news != p->s && p->s != 0)					\
    fprintf (stderr, "%s(): "						\
	     "Varování: změna počtu složek (z %d na %d).\n",		\
	     __func__, p->s, news);					\
  p->s = news;								\
}

SET_S (cold_p);
SET_S (dens_spline_p);
SET_S (dens_r4_p);
SET_S (diffus_equi_p);

#define SET_COMP_VALS(nazev_p, var, varslovy, typ) void			\
nazev_p##_set_##var (struct nazev_p *p, int s, typ *new_##var)		\
{									\
  if (s <= 0)								\
    {									\
      fprintf (stderr, "%s(): Počet složek má býti kladný!\n", __func__);\
      abort ();								\
    }									\
  else if (s != p->s && p->s != 0)					\
    fprintf (stderr, "%s(): Varování: "					\
	     "změna počtu složek (z %d na %d).\n", __func__, p->s, s);	\
  p->s = s;								\
  if (p->var)								\
    {									\
      fprintf (stderr, "%s(): "						\
	       "Varování: " #varslovy " již byly nastaveny.\n", __func__);\
      free (p->var);							\
    }									\
  p->var = malloc (s * sizeof (typ));					\
  memcpy (p->var, new_##var, (s * sizeof (typ)));			\
}

SET_COMP_VALS (cold_p, m, hmotnosti, double);
SET_COMP_VALS (cold_p, q, n  boje, double);
SET_COMP_VALS (diffus_equi_p, m, hmotnosti, double);
SET_COMP_VALS (dens_r4_p, d_1, hustoty  v  z  kladn   hladin  ,
	       double);
SET_COMP_VALS (diffus_equi_p, d_b, hustoty  v  z  kladn   hladin  ,
	       double);

#define KONSTANTA_DIPOL (mu_0/M_PI/4.)	/*_U_*/	/* μ_0/4π v SI, 1 v CGS */

rvec
dipolz (rvec * r, void *param)
{
  struct dipolz_p * /*restrict */ p = param;
  double ra = rvec_norma (r);
  rvec b;
  for (int i = 0; i < 3; i++)
    b.val[i] =
      -KONSTANTA_DIPOL * 3. * r->val[2] * p->moment * r->val[i] / (ra * ra *
								   ra * ra *
								   ra);
  b.val[2] += KONSTANTA_DIPOL * p->moment / (ra * ra * ra);
  return b;
}

// momentálně předpokládá, že na 0. místě jsou elektrony (ty to počítá
// z nábojové rovnováhy) a na dalších místech jsou kladně jednomocné ionty
// (ty to počítá z jejich difus. ekvilibria)
void
diffus_equi (double *dest, rvec * poloha, void *param)
{
  struct diffus_equi_p * /*restrict */ p = param;
  double r = rvec_norma (poloha);
  for (int i = 1; i < p->s; i++)
    dest[i] =
      p->d_b[i] * exp (p->gm * p->m[i] * (1. / r - 1. / p->b) / (k_B * p->t));
  dest[0] = 0;
  for (int i = 1; i < p->s; i++)
    dest[0] += dest[i];
}

void
dens_r4 (double *dest, rvec * poloha, void *param)
{
  struct dens_r4_p * /*restrict */ p = param;
  double r4 = rvec_in (poloha, poloha);
  r4 *= r4;
  for (int i = 0; i < p->s; i++)
    dest[i] = p->d_1[i] / r4;
}

void
dens_r4_clean (struct dens_r4_p *p)
{
  if (p->d_1)
    free (p->d_1);
}

void
dens_spline (double *dest, rvec * poloha, void *param)
{
  struct dens_spline_p * /*restrict */ p = param;
  double r = rvec_norma (poloha);
  int ec;
  for (int i = 0; i < p->s; i++)
    if (ec =
	gsl_interp_eval_e (p->funcs[i], p->xas[i], p->yas[i], r,
			   p->accs[i], dest + i))
      {
	fprintf (stderr, "%s: %d %s\n", __func__, ec, gsl_strerror (ec));
	abort ();
      }
}

int
dens_spline_initfun_cmp (const void *a, const void *b)	// pro qsort()
{
  if ((*((double (*)[2]) a))[0] < (*((double (*)[2]) b))[0])
    return -1;
  if ((*((double (*)[2]) a))[0] == (*((double (*)[2]) b))[0])
    return 0;
  return 1;
}

/* Nejsou tu v podstatě žádné kontroly na formát souboru: */
int
dens_spline_initfun (struct dens_spline_p *p, int no,
		     const char filename[], int colx, int coly)
{
  FILE *fr;
  if (!(fr = fopen (filename, "r")))
    {
      perror (__func__);
      abort ();
    }
  double *xval, *yval;
  int msize = 1024;
  xval = malloc (msize * sizeof (double));
  yval = malloc (msize * sizeof (double));
  double *val1, *val2;
  int col1, col2;
  if (colx > coly)
    col1 = coly, col2 = colx, val1 = yval, val2 = xval;
  if (colx < coly)
    col1 = colx, col2 = coly, val1 = xval, val2 = yval;
  if (colx == coly)
    {
      fprintf (stderr, "error: %s(): colx == coly!\n", __func__);
      abort ();
    }
  int c;
  int i = 0;
  while ((c = getc (fr)) != EOF)
    if (ungetc (c, fr), c == '#')
      dojedradek (fr);
    else if (najdisloupec (fr, col1 - 1))
      dojedradek (fr);
    else if (1 != fscanf (fr, "%lf", val1 + i))
      dojedradek (fr);
    else if (najdisloupec (fr, col2 - col1 - 1))
      dojedradek (fr);
    else if (1 != fscanf (fr, "%lf", val2 + i))
      dojedradek (fr);
    else
      {
	++i;
	if (i >= msize)
	  {
	    msize *= 2;
	    val1 = realloc (val1, msize * sizeof (double));
	    val2 = realloc (val2, msize * sizeof (double));
	  }
	dojedradek (fr);
      }
  fclose (fr);
  if (colx > coly)
    yval = val1, xval = val2;
  else
    yval = val2, xval = val1;
  for (int ii = 1; ii < i; ii++)
    if (xval[ii] <= xval[ii - 1])
      {				/* gsl_interp_alloc() vyžaduje ostře rostoucí xval */
	fprintf (stderr,
		 "%s(): sloupec r není ostře rostoucí; seřazuji.\n",
		 __func__);
	double tmp[i][2];
	for (int j = 0; j < i; j++)
	  tmp[j][0] = xval[j], tmp[j][1] = yval[j];
	qsort (tmp, i, sizeof (double[2]), dens_spline_initfun_cmp);
	xval[0] = tmp[0][0], yval[0] = tmp[0][1];
	int w = 1;
	for (int r = 1; r < i; r++)	// Zajištění unikátnosti
	  if (tmp[r][0] > tmp[r - 1][0])
	    xval[w] = tmp[r][0], yval[w] = tmp[r][1], w++;
	i = w;
	break;
      }

  p->funcs[no] = gsl_interp_alloc (gsl_interp_cspline, i);
  p->xas[no] = xval;
  p->yas[no] = yval;
  p->sizes[no] = i;
  gsl_interp_init (p->funcs[no], xval, yval, i);
  p->accs[no] = gsl_interp_accel_alloc ();
  return 0;
}

void
dens_spline_clean (struct dens_spline_p *p)
{
  for (int i = 0; i < p->s; i++)
    {
      gsl_interp_free (p->funcs[i]);
      gsl_interp_accel_free (p->accs[i]);
      free (p->xas[i]);
      free (p->yas[i]);
    }
  if (p->file)
    {
      free (p->file);
      p->file = NULL;
    }
  if (p->cols)
    {
      free (p->cols);
      p->cols = NULL;
    }
}
