#define dbg stderr

#ifndef IRI_DATA_PATH
#define IRI_DATA_PATH "iri-2012"
#endif

const char *iri_data_path = IRI_DATA_PATH;

#include "iridens.h"
#include "consts.h"
#include "fortrantypes.h"
#include <math.h>
#include <string.h>
#include <unistd.h>

#include <stdio.h>		//pro ladění
void mat_fprint (FILE *, mat *);	//taktéž
#include <assert.h>


void
iri (double *dest, rvec * poloha, void *param)
{
  const struct iri_p *p = param;
  struct iri_dens1 d1 = iri_getdens1 (*poloha);
  for (int i = 0; i < p->s; i++)
    {
      switch (p->mapping[i])
	{
	case 0:
	  dest[i] = d1.e;
	  break;
	case 1:
	  dest[i] = d1.o;
	  break;
	case 2:
	  dest[i] = d1.h;
	  break;
	case 3:
	  dest[i] = d1.he;
	  break;
	case 4:
	  dest[i] = d1.o2;
	  break;
	case 5:
	  dest[i] = d1.no;
	  break;
	case 6:
	  dest[i] = d1.n;
	  break;
	case 7:
	  dest[i] = d1.clus;
	  break;
	default:
	  fprintf (stderr, "%s: mapping[%d] == %d; toto nastat nemělo\n",
		   __func__, i, p->mapping[i]);
	  abort ();
	}
    }
}

const char *const iri_ion_names[IRIDENS_MAX_S] =
  { "e", "O", "H", "He", "O2", "NO", "N", "cluster" };

void iri_sub_ (integer (*jf)[50], integer * jmag, real * alati, real * along,
	       integer * iyyyy, integer * mmdd, real * dhour, real * heibeg,
	       real * heiend, real * heistp, real (*outf)[1000][20],
	       real (*oarr)[50]);

void iri_web_ (integer * jmag, integer (*jf)[50], real * alati, real * along,
	       integer * iyyyy, integer * mmdd, integer * iut, real * dhour,
	       real * height, integer * h_tec_max, integer * ivar,
	       real * vbeg, real * vend, real * vstp, real (*a)[1000][20],
	       real (*b)[1000][100]);

void initialize_ ();

// Něco z toho možná nevyužiji
void feldg_ (real * glat, real * glon, real * alt, real * bnorth,
	     real * beast, real * bdown, real * babs);

void feldcof_ (real * year, real * dimo);

void igrf_sub_ (real * xlat, real * xlong, real * year, real * height,
		real * xl, integer * icode, real * dip, real * dec);



integer iri_initialized = 0;
int iriwrap_iyyyy = 2011, iriwrap_mmdd = -180;
real iriwrap_dhour = 30;
integer iriwrap_jf[50] = { 1, 1, 1, 0, 0,
  0, 1, 1, 1, 1,
  1, 0, 1, 1, 1,
  1, 1, 1, 1, 1,
  0 /*? */ , 0 /* sic! JF(22) */ , 0, 1, 1,
  1, 1, 1, 0, 0,
  1, 1, 0, 1, 0
    /* 36 – 50 nevyužito v IRI 2012 */
};

/* Zatím takhle, příště by měl mít čas v parametrech místo spoléhání na stat */
/* Všude momentálně používám geografické souřadnice; osa y míří na 0 v.d. */
struct iri_dens1
iri_getdens1 (rvec xyz)
{
  if (!iri_initialized)
    {
      if (chdir (IRI_DATA_PATH))
	{
	  perror (__func__);
	  fprintf (stderr,
		   "%s: Nelze přejít do adresáře %s s daty IRI.\n",
		   __func__, iri_data_path);
	  abort ();
	}
      initialize_ ();
      iri_initialized = 1;
    }
  double ghei = 0, glat, glong;	// geograf. souř.
  for (int i = 0; i < 3; i++)
    ghei += SQ (xyz.val[i]);
  ghei = (sqrt (ghei) - RADIUS_TERRAE) / KILOMETR;	// Převedeno na km /*_U_*/
  glong = atan2 (xyz.val[0], xyz.val[1]) * 180 / M_PI;
  glat =
    atan (xyz.val[2] / sqrt (SQ (xyz.val[0]) + SQ (xyz.val[1]))) * 180 / M_PI;
#ifdef DEBUG
  fprintf (dbg, "%s: x = %e m, y = %e m, z = %e m;\n", __func__,
	   xyz.val[0], xyz.val[1], xyz.val[2]);
  fprintf (dbg, "%s: height = %e km, lat = %e deg, long = %e deg.\n",
	   __func__, ghei, glat, glong);
#endif
//  iriwrap_jf[21] = 0; // Chceme to v m-3, ne v % (už to je v defaults)
  integer jmag = 0;		//const
  real alati = glat, along = glong, heibeg = ghei, heiend = ghei, heistp =
    0.1f;
  static real outf[1000][20], oarr[50];
  iri_sub_ (&iriwrap_jf, &jmag, &alati, &along, &iriwrap_iyyyy, &iriwrap_mmdd,
	    &iriwrap_dhour, &heibeg, &heiend, &heistp, &outf, &oarr);
  struct iri_dens1 d;
  d.e = outf[0][0];
  d.o = outf[0][4];
  d.h = outf[0][5];
  d.he = outf[0][6];
  d.o2 = outf[0][7];
  d.no = outf[0][8];
  d.clus = outf[0][9];
  d.n = outf[0][10];
  return d;
}

// Bude toto třeba? (Nenastavuje to iri automaticky taky?)
// Nebude, nastavuje, můžeme to smazat.
void
igrf_setyear (real decyear)
{
  real dimo;			// Lieutenant
  feldcof_ (&decyear, &dimo);
}

// Duplicitní kód, v mag-igrf je to chytřeji
rvec
igrf_b (rvec xyz, double *b_abs /* Volitelný výstup, jinak NULL */ )
{
  rvec dolu;
  dolu = rvec_normuj (&xyz);
  dolu = rvec_mul_con (-1, &dolu);
#ifdef DEBUG
  fprintf (dbg, "igrf_b: xyz = (%e %e %e) cm\n", xyz.val[0], xyz.val[1],
	   xyz.val[2]);
  fprintf (dbg, "igrf_b: dolu = (%e %e %e)\n", dolu.val[0], dolu.val[1],
	   dolu.val[2]);
#endif

  double ghei = 0, glati, glong;	// geograf. souř.
  for (int i = 0; i < 3; i++)
    ghei += SQ (xyz.val[i]);
  ghei = sqrt ((ghei) - RADIUS_TERRAE) / KILOMETR;	// Převedeno na km
  glong = atan2 (xyz.val[0], xyz.val[1]);
  real glongdeg = glong * 180 / M_PI;
  glati = atan (xyz.val[2] / sqrt (SQ (xyz.val[0]) + SQ (xyz.val[1])));
  real glatdeg = glati * 180 / M_PI;

  real alt = ghei, bnorth, beast, bdown, babs;
  feldg_ (&glatdeg, &glongdeg, &alt, &bnorth, &beast, &bdown, &babs);
  /* Údajně je to již v gaussech */
  rvec b;
  b.val[0] = (beast * cos (glong) - bnorth * sin (glati) * sin (glong) +
	      bdown * dolu.val[0]) * GAUSS;
  b.val[1] =
    (-beast * sin (glong) - bnorth * sin (glati) * cos (glong) +
     bdown * dolu.val[1]) * GAUSS;
  b.val[2] = (bnorth * cos (glati) + bdown * dolu.val[2]) * GAUSS;
  if (b_abs)
    *b_abs = babs * GAUSS;
  return b;
}

int
iri_settime (integer year, integer day, real hour)
{
  iriwrap_iyyyy = year;
  iriwrap_mmdd = -day;
  iriwrap_dhour = hour + 25;
}

/* 
 * Duplikát stejného makra v cold.c; FIXME dát taková makra do extra hlavičkového
 * souboru
 */

#define SET_S(nazev_p) void 						\
	nazev_p##_set_s (struct nazev_p *p, int news) 				\
{ 									\
	  if (news <= 0) 							\
	    { 									\
		          fprintf (stderr, "%s(): Počet složek má býti kladný!\n",__func__);\
		          abort ();								\
		        }									\
	  else if (news != p->s && p->s != 0)					\
	    fprintf (stderr, "%s(): "						\
			    	     "Varování: změna počtu složek (z %d na %d).\n",		\
			    	     __func__, p->s, news);					\
	  p->s = news;								\
}

SET_S (iri_p);
