#include "igrf.h"
#include "fortrantypes.h"
#include <math.h>
#ifndef IGRF_DATA_PATH
#define IGRF_DATA_PATH "./iri-2012"
#endif
const char *igrf_data_path = IGRF_DATA_PATH;

// podrutiny z igrf.for
/* // Používám nížeúrovňovou a užitečnější rutinu feldg_
int igrf_sub_(const real *xlat, const real *xlong, const real *year, const real *height,
		real *xl, integer *icode, real *dipl, real *babs);
*/

// SUBROUTINE FELDG(GLAT,GLON,ALT,BNORTH,BEAST,BDOWN,BABS)           
// C CALCULATES EARTH MAGNETIC FIELD FROM SPHERICAL HARMONICS MODEL
//
// C  INPUT:  ENTRY POINT FELDG
// C               GLAT  GEODETIC LATITUDE IN DEGREES (NORTH)
// C               GLON  GEODETIC LONGITUDE IN DEGREES (EAST)
// C               ALT   ALTITUDE IN KM ABOVE SEA LEVEL
//
// C  OUTPUT: BABS   MAGNETIC FIELD STRENGTH IN GAUSS
// C          BNORTH, BEAST, BDOWN   COMPONENTS OF THE FIELD WITH RESPECT
// C                 TO THE LOCAL GEODETIC COORDINATE SYSTEM, WITH AXIS
// C                 POINTING IN THE TANGENTIAL PLANE TO THE NORTH, EAST
// C                 AND DOWNWARD.   
int feldg_ (const real * glat, const real * glon, const real * alt,
	    real * bnorth, real * beast, real * bdown, real * babs);
// C          ENTRY POINT FELDC
// C               V(3)  CARTESIAN COORDINATES IN EARTH RADII (6371.2 KM)
// C                       X-AXIS POINTING TO EQUATOR AT 0 LONGITUDE
// C                       Y-AXIS POINTING TO EQUATOR AT 90 LONG.
// C                       Z-AXIS POINTING TO NORTH POLE
int feldc_ (const real * v, real * b);

// SUBROUTINE FELDCOF(YEAR,DIMO)
// C------------------------------------------------------------------------
// C  DETERMINES COEFFICIENTS AND DIPOL MOMENT FROM IGRF MODELS
// C
// C       INPUT:  YEAR    DECIMAL YEAR FOR WHICH GEOMAGNETIC FIELD IS TO
// C                       BE CALCULATED
// C       OUTPUT: DIMO    GEOMAGNETIC DIPOL MOMENT IN GAUSS (NORMALIZED 
// C                       TO EARTH'S RADIUS) AT THE TIME (YEAR)
int feldcof_ (const real * year, real * dim0);

// initize_() je nutno volat před čímkoliv dalším.
// SUBROUTINE INITIZE
int initize_ (void);

struct igrf_p *igrf_owner = NULL;

// transformace souřadnic
// Osa y míří na 0 v.d.
#define DG (M_PI / 180.)
#define INVDG (180. / M_PI)
#define SQ(x) ((x)*(x))

#if 0				// Použijeme entry point feldc_
inline void
rvec2geod_dg_km (const rvec * r, float *lat_dg, float *long_dg, float *hei_km)
{
  double norma = rvec_norma (r);
  *lat_dg = asinf (r->val[2] / norma) * INVDG;
  *long_dg = atanf (r->val[0] / r->val[1]) * INVDG;
  if (r->val[1] < 0)
    {
      if (r->val[0] < 0)
	*long_dg -= 180.;
      else
	*long_dg += 180.;
    }
  if (r->val[1] == 0 && r->val[1] > 0)
    *long_dg = 90.;
  if (r->val[1] == 0 && r->val[1] < 0)
    *long_dg = -90.;
  *hei_km = (norma - RADIUS_TERRAE) / KILOMETR /*_U_*/ ;
}
#endif

#include <unistd.h>

int
igrf_checkinit (struct igrf_p *p)
{
  if (p == igrf_owner)
    return 0;
  /* TODO testy na rozsah year */
  if (p->initized && p->initized != p)
    goto igrf_checkinit_weirderr;
  if (!p->initized && !igrf_owner)
    goto igrf_checkinit_init;
  if (p->initized && !igrf_owner)
    goto igrf_checkinit_weirderr;
  if (p->initized != igrf_owner)
    goto igrf_checkinit_changeowner;
  char oldpath[BUFSIZ];
igrf_checkinit_changeowner:
  fprintf (stderr,
	   "%s: IGRF již bylo inicializováno s rokem %f, měním na %f\n",
	   __func__, (double) (igrf_owner->year), (double) (p->year));
igrf_checkinit_init:
  if (NULL == getcwd (oldpath, BUFSIZ))
    goto igrf_checkinit_weirderr;
  if (chdir (igrf_data_path))
    {
      perror (__func__);
      abort ();
    }
  initize_ ();
  p->initized = igrf_owner = p;
  feldcof_ (&(p->year), &(p->dim0));
  if (chdir (oldpath))
    {
      perror (__func__);
      goto igrf_checkinit_weirderr;
    }
  return 1;
igrf_checkinit_weirderr:
  fprintf (stderr, "%s: Došlo k chybě, jež neměla nastat.\n", __func__);
  abort ();
}

rvec
igrf (rvec * r, void *param)
{
  igrf_checkinit (param);
  real v[3], b[3];
  v[0] = r->val[1] / RADIUS_TERRAE;	// sic! prohozené osy!
  v[1] = r->val[0] / RADIUS_TERRAE;
  v[2] = r->val[2] / RADIUS_TERRAE;
  feldc_ (v, b);
  rvec vysl;
  vysl.val[0] = b[1] * GAUSS;	// opět prohozené osy! /*_U_*/
  vysl.val[1] = b[0] * GAUSS;
  vysl.val[2] = b[2] * GAUSS;
  return vysl;
}
