
/* deskriptory pro ladicí a chybový výstup */
#define dbg stderr
#define _err_out stderr

#include "matvec.h"

#ifndef DEBUG
#define NDEBUG
#endif

#define SQ(x) ((x)*(x))

/* operace s maticemi; možná půjde do extra souboru */
mat unit = { {{1, 0, 0}
	      , {0, 1, 0}
	      , {0, 0, 1}
	      }
};

mat
mat_tr (mat * a)
{
  mat m;
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      m.val[i][j] = a->val[j][i];
  return m;
}

mat
mat_add (mat * a, mat * b)
{
  mat m;
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      m.val[i][j] = a->val[i][j] + b->val[i][j];
  return m;
}

mat
mat_mul (mat * b, mat * a)
{
  mat m;
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      {
	double complex v = 0;
	for (int k = 0; k < 3; k++)
	  v += a->val[i][k] * b->val[k][j];
	m.val[i][j] = v;
      }
  return m;
}

/* skalární součin */
double
rvec_in (rvec * a, rvec * b)
{
  double x = 0;
  for (int i = 0; i < 3; i++)
    x += a->val[i] * b->val[i];
  return x;
}

mat
mat_mul_con (double complex a, mat * b)
{
  mat m;
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      m.val[i][j] = a * b->val[i][j];
  return m;
}

rvec
rvec_mul_con (double a, rvec * b)
{
  rvec v;
  for (int i = 0; i < 3; i++)
    v.val[i] = a * b->val[i];
  return v;
}

/*  V kódu je to zatím všelijak, pro nový použít toto */
double
rvec_norma (rvec * v)
{
  double n = 0;
  for (int i = 0; i < 3; i++)
    n += SQ (v->val[i]);
  n = sqrt (n);
  return n;
}

double
rvec_dist (rvec * x, rvec * y)
{
  double n = 0;
  for (int i = 0; i < 3; i++)
    n += SQ (x->val[i] - y->val[i]);
  n = sqrt (n);
  return n;
}

rvec
rvec_normuj (rvec * v)
{
  double norma = 0;
  for (int i = 0; i < 3; i++)
    norma += SQ (v->val[i]);
  norma = sqrt (norma);
  rvec nv;
  for (int i = 0; i < 3; i++)
    nv.val[i] = v->val[i] / norma;
  return nv;
}

rvec
rvec_vp (rvec * a, rvec * b)
{
  rvec x;
  x.val[0] = a->val[1] * b->val[2] - a->val[2] * b->val[1];
  x.val[1] = a->val[2] * b->val[0] - a->val[0] * b->val[2];
  x.val[2] = a->val[0] * b->val[1] - a->val[1] * b->val[0];
  return x;
}

rvec
rvec_add (rvec * a, rvec * b)
{
  rvec n;
  for (int i = 0; i < 3; i++)
    n.val[i] = a->val[i] + b->val[i];
  return n;
}

rvec
rvec_sub (rvec * a, rvec * b)
{
  rvec n;
  for (int i = 0; i < 3; i++)
    n.val[i] = a->val[i] - b->val[i];
  return n;
}

void
mat_fprint (FILE * f, mat * m)
{
  fprintf (f, "%e + %ei   %e + %ei   %e + %ei\n", creal (m->val[0][0]),
	   cimag (m->val[0][0]), creal (m->val[1][0]), cimag (m->val[1][0]),
	   creal (m->val[2][0]), cimag (m->val[2][0]));
  fprintf (f, "%e + %ei   %e + %ei   %e + %ei\n", creal (m->val[0][1]),
	   cimag (m->val[0][1]), creal (m->val[1][1]), cimag (m->val[1][1]),
	   creal (m->val[2][1]), cimag (m->val[2][1]));
  fprintf (f, "%e + %ei   %e + %ei   %e + %ei\n", creal (m->val[0][2]),
	   cimag (m->val[0][2]), creal (m->val[1][2]), cimag (m->val[1][2]),
	   creal (m->val[2][2]), cimag (m->val[2][2]));
}

void
mat_print (mat * m)
{
  mat_fprint (stdout, m);
}

complex double
det (mat * m)
{
  return m->val[0][0] * m->val[1][1] * m->val[2][2]
    - m->val[2][0] * m->val[1][1] * m->val[0][2]
    + m->val[0][1] * m->val[1][2] * m->val[2][0]
    - m->val[2][1] * m->val[1][2] * m->val[0][0]
    + m->val[0][2] * m->val[1][0] * m->val[2][1]
    - m->val[2][2] * m->val[1][0] * m->val[0][1];
}

mat
kxkx (rvec k)
{
  mat m = { {{-k.val[1] * k.val[1] - k.val[2] * k.val[2], k.val[0] * k.val[1],
	      k.val[0] * k.val[2]}
	     , {k.val[0] * k.val[1],
		-k.val[2] * k.val[2] - k.val[0] * k.val[0],
		k.val[1] * k.val[2]}
	     , {k.val[0] * k.val[2], k.val[1] * k.val[2],
		-k.val[0] * k.val[0] - k.val[1] * k.val[1]}
	     }
  };
#ifdef DEBUG_G
  fprintf (dbg, "matice kxkx = \n");
  mat_fprint (dbg, &m);
#endif
  return m;
}

/* Orthogonální transformace, jež osu z převede do daného směru
 * a druhý vektor volitelně vytyčuje rovinu xz. Určeno k transformaci
 * matice susceptibility z tvaru, kdy osa z splývá se směrem magnetické
 * indukce, k obecnému směru B.
 *
 * Aby bylo jasno: Pokud χ_B je matice susceptibility, kde třetí rozměr
 * je vztažen ke směru B, a M je matice vyšedší z této funkce,
 * pak susceptibilita χ v "normální" basi xyz je
 * χ = M' χ_B M
 */
#define SCAL_PROD(vec1, vec2) \
	((vec1).val[0] * (vec2).val[0] + \
	 (vec1).val[1] * (vec2).val[1] + \
	 (vec1).val[2] * (vec2).val[2])
mat
ortho_xz (rvec * nz, rvec * xspan)
{
  // | u2 u3 u1(=*nz) |
  mat m;
  rvec u1 = rvec_normuj (nz);
  for (int i = 0; i < 3; i++)
    m.val[i][2] = u1.val[i];
  rvec xs, u2, u3;
  double uv = 1.;
  if (xspan)
    {
      xs = rvec_normuj (xspan);
      uv = SCAL_PROD (xs, u1);
    }
  if (0.99999 > fabs (uv))	// Případ rozumného úhlu, orthogonalizuj
    for (int i = 0; i < 3; i++)
      u2.val[i] = xs.val[i] - uv * u1.val[i];
  else
    {				// vyber osu, na niž je nz "nejkolmější" 
      double min = fabs (u1.val[0]);
      int imin = 0;
      if (min > fabs (u1.val[1]))
	min = fabs (u1.val[1]), imin = 1;
      if (min > fabs (u1.val[2]))
	imin = 2;
      for (int i = 0; i < 3; i++)
	u2.val[i] = ((i == imin) ? 1 : 0) - u1.val[imin] * u1.val[i];
    }
  u2 = rvec_normuj (&u2);
  // Vektor y určíme z vnějšího součinu
  u3.val[0] = u1.val[1] * u2.val[2] - u2.val[1] * u1.val[2];
  u3.val[1] = u1.val[2] * u2.val[0] - u2.val[2] * u1.val[0];
  u3.val[2] = u1.val[0] * u2.val[1] - u2.val[0] * u1.val[1];

  for (int i = 0; i < 3; i++)
    {
      m.val[i][0] = u2.val[i];
      m.val[i][1] = u3.val[i];
    }
#ifdef DEBUG
  fprintf (dbg, "ortho_xz (%e, %e, %e) = \n", nz->val[0], nz->val[1],
	   nz->val[2]);
  mat_fprint (dbg, &m);
#endif
  return m;
}
