/* deskriptory pro ladicí a chybový výstup */
#define dbg stderr
#define _err_out stderr

#include "trace.h"
#include <gsl/gsl_deriv.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_min.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_odeiv2.h>
#include <stdio.h>

#ifndef DEBUG
#define NDEBUG
#endif
#include <assert.h>
#include "consts.h"

/* Dispersní vztah g (r, k, omega) == 0 */
double complex
g_std (rvec r, rvec k, double complex omega, susc_t susc, void *susc_p)
{
  mat m = susc (&r, &k, omega, susc_p);
#ifdef DEBUG_G
  fprintf (dbg, "χ = \n");
  mat_fprint (dbg, &m);
#endif
  m = mat_add (&unit, &m);
  m = mat_mul_con (omega * omega / (C * C), &m);
#ifdef DEBUG_G
  fprintf (dbg, "ε * μ_0 * ω**2 = \n");
  mat_fprint (dbg, &m);
#endif
  mat kk = kxkx (k);
  m = mat_add (&m, &kk);
#ifdef DEBUG_G
  fprintf (dbg, "matice G = \n");
  mat_fprint (dbg, &m);
#endif
  return det (&m);
}

// Bude-li nutno funkci znásilnit (pro whamp)
double complex
  (*g) (rvec r, rvec k, double complex omega, susc_t susc, void *susc_p) =
  g_std;

/* Výpočet g(r, c*k, omega) */
double
g_kmag (double c, void *param)
{
  struct g_par p = *((struct g_par *) param);
  for (int i = 0; i < 3; i++)
    p.k.val[i] *= c;
  double vysl = creal (g (p.r, p.k, p.omega, p.susc, p.susc_p));
#ifdef DEBUG
  fprintf (dbg, "g_kmag: g(c = %.16e) = %e\n", c, vysl);
#endif
  return vysl;
}

/* (Skoro) totéž v absolutní hodnotě */
double
g_abs_kmag (double c, void *param)
{
  struct g_par p = *((struct g_par *) param);
  for (int i = 0; i < 3; i++)
    p.k.val[i] *= c;
  double vysl = cabs (g (p.r, p.k, p.omega, p.susc, p.susc_p));
#ifdef DEBUG
  fprintf (dbg, "g_abs_kmag: |g(c = %.16e)| = %e\n", c, vysl);
#endif
  return vysl;
}



/* Hnus. Aby gsl mohla počítat parciální derivace g(): */
/* Možná to půjde umravnit pomocí maker */
/* Následující funkce počítají hodnotu g z dané struct g_par
 * kromě jedné proměnné, která je dána v prvním argumentu
 */

double
gomega (double omega, void *param)
{
  struct g_par *p = ((struct g_par *) param);
  double complex gv = g (p->r, p->k, omega, p->susc, p->susc_p);
  return creal (gv);
}

double
gr0 (double r0, void *param)
{
  struct g_par *p = ((struct g_par *) param);
  rvec r = p->r;
  r.val[0] = r0;
  double complex gv = g (r, p->k, p->omega, p->susc, p->susc_p);
  return creal (gv);
}

double
gr1 (double r1, void *param)
{
  struct g_par *p = ((struct g_par *) param);
  rvec r = p->r;
  r.val[1] = r1;
  double complex gv = g (r, p->k, p->omega, p->susc, p->susc_p);
  return creal (gv);
}

double
gr2 (double r2, void *param)
{
  struct g_par *p = ((struct g_par *) param);
  rvec r = p->r;
  r.val[2] = r2;
  double complex gv = g (r, p->k, p->omega, p->susc, p->susc_p);
  return creal (gv);
}

double
gk0 (double k0, void *param)
{
  struct g_par *p = ((struct g_par *) param);
  rvec k = p->k;
  k.val[0] = k0;
  double complex gv = g (p->r, k, p->omega, p->susc, p->susc_p);
  return creal (gv);
}

double
gk1 (double k1, void *param)
{
  struct g_par *p = ((struct g_par *) param);
  rvec k = p->k;
  k.val[1] = k1;
  double complex gv = g (p->r, k, p->omega, p->susc, p->susc_p);
  return creal (gv);
}

double
gk2 (double k2, void *param)
{
  struct g_par *p = ((struct g_par *) param);
  rvec k = p->k;
  k.val[2] = k2;
  double complex gv = g (p->r, k, p->omega, p->susc, p->susc_p);
  return creal (gv);
}

/* konec hnusu */

/*
 * Funkce vypočítá dk/dt a dr/dt pro daná k, r, omega
 * navíc aktualisuje strukturu s parametry (sic!)
 */

/* 1e-25 = cca polovina přesnosti double */
//TODO ošetřit chyby
//#define DKROK (1./0x2000000)
double DKROK = (1. / 0x2000);
int
derivace (double t, const double y[], double dydt[],
	  void *param /* vše je v t a y[] */ )
	/* y == {r[0], r[1], r[2], k[0], k[1], k[2]}  */
{
  struct g_par *p = (struct g_par *) param;
  /* !!! aktualisace předané struktury *param !!! */
  for (int i = 0; i < 3; i++)
    p->r.val[i] = y[i];
  for (int i = 0; i < 3; i++)
    p->k.val[i] = y[i + 3];
  double dgdomega, dgdk[3], dgdr[3], dgdomega_err, dgdk_err[3], dgdr_err[3];
  gsl_function F;
  F.params = p;

  double knorma = rvec_norma (&(p->k)), rnorma = rvec_norma (&(p->r));

  F.function = gomega;
  gsl_deriv_central (&F, p->omega, DKROK * p->omega, &dgdomega,
		     &dgdomega_err);
  F.function = gk0;
  gsl_deriv_central (&F, p->k.val[0], DKROK * knorma, dgdk + 0, dgdk_err + 0);
  F.function = gk1;
  gsl_deriv_central (&F, p->k.val[1], DKROK * knorma, dgdk + 1, dgdk_err + 1);
  F.function = gk2;
  gsl_deriv_central (&F, p->k.val[2], DKROK * knorma, dgdk + 2, dgdk_err + 2);
  F.function = gr0;
  /* DKROK * rnorma pravděpodobně nebude zcela ideálním krokem; jeden by čekal,
   * že na aktuální poloze to záviset nebude, tedy alespon ne takovýmto
   * způsobem... */
  gsl_deriv_central (&F, p->r.val[0], DKROK /* * rnorma */ , dgdr + 0,
		     dgdr_err + 0);
  F.function = gr1;
  gsl_deriv_central (&F, p->r.val[1], DKROK /* * rnorma */ , dgdr + 1,
		     dgdr_err + 1);
  F.function = gr2;
  gsl_deriv_central (&F, p->r.val[2], DKROK /* * rnorma */ , dgdr + 2,
		     dgdr_err + 2);

  {
    double domegadk[3], domegadr[3];
    for (int i = 0; i < 3; i++)
      domegadk[i] = -dgdk[i] / dgdomega;
    for (int i = 0; i < 3; i++)
      domegadr[i] = -dgdr[i] / dgdomega;
/* omega / (k . grad_k omega */
    double fac = 1;
    //pro "ray velocity" by to bylo toto: //
    //p->omega / (p->k.val[0] * domegadk[0] +
    //        p->k.val[2] * domegadk[1] + p->k.val[2] * domegadk[2]);

    for (int i = 0; i < 3; i++)
      dydt[i] = domegadk[i] * fac;
    for (int i = 0; i < 3; i++)
      dydt[i + 3] = -domegadr[i] * fac;
  }
  return GSL_SUCCESS;
}

/* 
 * Určení horní meze pro hledání kořenů 
 * Horní mez je přibližně tam, kde velikost g() klesne
 * pod velikost na počátku (determinant permitivity).
 *
 * TODO: Funkce možná není moc spolehlivá (práh někdy nastavuje
 * až pod největší kořen)
 *
 * Je otázkou, zda není funkce zbytečná, třeba by to minimalisační
 * algoritmus zvládl sám se srovnatelnou effektivitou
 */

//#define MOC 3.40282347e+38
#define MOC (omega * 1e8 / C)
double
interval (rvec r, rvec smer, double omega,
	  susc_t susceptibilita, void *susc_p)
{
  rvec k = { {0, 0, 0} };
  double thr = cabs (g (r, k, omega, susceptibilita, susc_p));
#ifdef DEBUG
  fprintf (dbg, "interval: g(|k| = 0) = %e\n", thr);
#endif /* DEBUG */
  double norma =
    sqrt (SQ (smer.val[0]) + SQ (smer.val[1]) + SQ (smer.val[2]));
  for (int i = 0; i < 3; i++)
    smer.val[i] /= norma;
  for (int i = 0; i < 3; i++)
    k.val[i] = smer.val[i] * MOC;
  norma = MOC;
  double ca;
  int sign = (creal (g (r, k, omega, susceptibilita, susc_p)) > 0) ? 1 : -1;
//  /*smazat, pro ladění */ double temp;
//  temp = creal (g (r, k, omega, susceptibilita, susc_p));
  while (((ca =
	   sign * creal (g (r, k, omega, susceptibilita, susc_p))) > thr))
    {
#ifdef DEBUG
      fprintf (dbg, "interval: g(|k| = %e) = %e\n", norma, ca);
#endif /* DEBUG */
      for (int i = 0; i < 3; i++)
	k.val[i] /= 2.;
      norma /= 2;
      /*smazat/////// */
//      temp = creal (g (r, k, omega, susceptibilita, susc_p));
    }
  return 4 * norma;
}

/* 
 * (Snad) nalezne JEDNO řešení disp. relace ve směru k
 * Nevím, jak je to stabilní pro jiné než studené plasma
 * (Předpokládá se, že existuje pouze jeden extrém mezi kořeny)
 */

/* Relativní přesnost hledání kořene, vzhledem ke g(k = 0) */
//#define R_ACC 1e-6
double R_ACC = 1e-6;

double
groot1 (rvec r, rvec smer, double omega, susc_t susc,
	void *susc_p, double *err)
{
  int iter = 0, max_iter = 1000;
  smer = rvec_normuj (&smer);
  double a = 0, b = interval (r, smer, omega, susc, susc_p);
  double m;

  gsl_function F;
  F.function = g_abs_kmag;
  struct g_par p = { r, smer, omega, susc, susc_p };
  F.params = &p;

  double g_v_0 = g_abs_kmag (0, &p);
#ifdef DEBUG
  fprintf (dbg, "|g| v nule = %e\n", g_v_0);
#endif /* DEBUG */

  {
    double g_v_b = g_abs_kmag (b, &p);
    m = b * 0.999999999999;
    double g_v_m;
    for (int i = 0; i < 100; i++)
      {
	if ((g_v_m = g_abs_kmag (m, &p)) < g_v_0 && g_v_m < g_v_b)
	  goto odhad_nalezen;
	else
	  m /= 2;
      }
    fprintf (_err_out, "groot1: odhad nenalezen\n");
    abort ();
  }

  const gsl_min_fminimizer_type *T;
  gsl_min_fminimizer *s;
odhad_nalezen:
  T = gsl_min_fminimizer_brent;
  s = gsl_min_fminimizer_alloc (T);

  int status;
  status = gsl_min_fminimizer_set (s, &F, m, a, b);

  double g_min;

  double oldm, olda, oldb;
  do
    {
      iter++;
      status = gsl_min_fminimizer_iterate (s);
      m = gsl_min_fminimizer_x_minimum (s);
      a = gsl_min_fminimizer_x_lower (s);
      b = gsl_min_fminimizer_x_upper (s);

      status = gsl_root_test_delta (a, b, 0, R_ACC);

      if (status == GSL_SUCCESS &&	/* Dodatečná podmínka i na velikost G */
	  ((g_min = gsl_min_fminimizer_f_minimum (s)) > g_v_0 * R_ACC))
	status = GSL_CONTINUE;

      /* Ošklivý hack: nic se nezměnilo */
      if (status == GSL_CONTINUE && olda == a && oldb == b && oldm == m)
	status = GSL_SUCCESS;

      if (status == GSL_SUCCESS)
	{
	  /* Hurá */
#ifdef DEBUG
	  fprintf (dbg, "groot1: Zkonvergovalo po %d. iteraci, |k| = %e\n"
		   "groot1: (a, b) = (%e, %e); g(|k| = m) = %e \n",
		   iter, m, a, b, g_min);
#endif /* DEBUG */
	  gsl_min_fminimizer_free (s);
	  if (err)
	    *err = b - a;
	  return m;
	}
      oldm = m, olda = a, oldb = b;
    }
  while (status == GSL_CONTINUE && iter < max_iter);

  if (status == GSL_CONTINUE)
    fprintf (_err_out, "groot1: NEzkonvergovalo po %d. iteraci:\n"
	     "groot1: m = %e; (a, b) = (%e, %e); g(|k| = m) = %e \n",
	     iter, m, a, b, g_min);
  else
    fprintf (_err_out, "groot1: Chyba %d při %d. iteraci:\n"
	     "groot1: m = %e, (a, b) = (%e, %e); g(|k| = m) = %e \n",
	     status, iter, m, a, b, g_min);
  gsl_min_fminimizer_free (s);
  if (err)
    *err = b - a;
  return m;
}

/* Určení velikosti vlnového vektoru pro daný směr 
 * 
 * První argument určuje, zda se má najít řešení pro nejmenší
 * |k| (vnitrni == 1) nebo pro větší |k| (vnitrni == -1).
 * Tiše se předpokládá, že funkce má nejvýš dva kořeny.
 *
 * TODO vytvořit zkratku, pokud groot1 rovnou najde resonanci
 * Bůh ví, co to ted vlastně dělá.
 */

double
kmag (int vnitrni, rvec r, rvec smer, double omega,
      susc_t susc, void *susc_p, double *err)
{
  smer = rvec_normuj (&smer);
  double groot1_err;
  double mag = groot1 (r, smer, omega, susc, susc_p, &groot1_err);
  if (err)
    *err = groot1_err;

  /* 
   * Pokud má derivace v nalezeném kořeni stejné znaménko jako
   * hodnota v "nekonečnu", je to vnější kořen...
   */

  struct g_par p = { r, smer, omega, susc, susc_p };
  gsl_function F = { g_kmag, &p };
  double deriv, deriv_err;
  gsl_deriv_central (&F, mag, DKROK * mag, &deriv, &deriv_err);
  double g_inf = g_kmag (MOC, &p);
  double g_abs_v_0 = g_abs_kmag (0, &p);
#ifdef DEBUG
  fprintf (dbg, "kmag: dg/dk_r = %e +- %e\n", deriv, deriv_err);
  fprintf (dbg, "kmag: g(|k|=MOC) = %e\n", g_inf);
  fprintf (dbg, "kmag: g(0)  = %e\n", g_kmag (0., &p));
#endif

  /* Máme-li lichý počet kořenů, doufáme, že je jediný... */
  if (g_kmag (0., &p) * g_inf < 0.)
    return mag;

  double dolnimez, hornimez;

  if (deriv * g_inf < 0.)	/* Nalezen vnitřní kořen */
    if (vnitrni == 1)
      return mag;
    else
      {
	dolnimez = mag + 2 * groot1_err;	/* Vymyslet lépe... */
	hornimez = interval (r, smer, omega, susc, susc_p);
	//assert (g_kmag (dolnimez, &p) * g_kmag (hornimez, &p) < 0.);
	// Pokud k tomu dojde, nejspíš se jedná o dvojnásobný kořen:
	if (g_kmag (dolnimez, &p) * g_kmag (hornimez, &p) >= 0.)
	  goto dvojnasobny_koren;
      }
  else if (vnitrni == -1)
    return mag;
  else
    {
      dolnimez = 0.;
      hornimez = mag - 2 * groot1_err;	/* Vymyslet lépe... */
      //assert (g_kmag (dolnimez, &p) * g_kmag (hornimez, &p) < 0.);
      if (g_kmag (dolnimez, &p) * g_kmag (hornimez, &p) >= 0.)
	goto dvojnasobny_koren;
    }

  /* Netrefili jsme se, hledáme druhý kořen... */
  int status;
  int iter = 0, max_iter = 1000;
  const gsl_root_fsolver_type *T = gsl_root_fsolver_brent;
  gsl_root_fsolver *s = gsl_root_fsolver_alloc (T);
  gsl_root_fsolver_set (s, &F, dolnimez, hornimez);

  do
    {
      iter++;
      status = gsl_root_fsolver_iterate (s);
      mag = gsl_root_fsolver_root (s);
      dolnimez = gsl_root_fsolver_x_lower (s);
      hornimez = gsl_root_fsolver_x_upper (s);

      status = gsl_root_test_delta (hornimez, dolnimez, 0, R_ACC);
      if (status == GSL_SUCCESS &&	/* Dodatečná podmínka i na velikost G */
	  (g_abs_kmag (mag, &p) > R_ACC * g_abs_v_0))
	status = GSL_CONTINUE;

      if (status == GSL_SUCCESS)
	{
#ifdef DEBUG
	  fprintf (dbg, "kmag: Zkonvergovalo po %d. iteraci; |k| = %e\n",
		   iter, mag);
#endif
	  gsl_root_fsolver_free (s);
	  if (err)
	    *err = hornimez - dolnimez;
	  return mag;
	}
    }
  while (status == GSL_CONTINUE && iter < max_iter);
  gsl_root_fsolver_free (s);
  fprintf (_err_out, "kmag: Nezkonvergovalo po %d. iteraci; |k| = %e\n", iter,
	   mag);
  if (err)
    *err = hornimez - dolnimez;
  return mag;

dvojnasobny_koren:
  fprintf (_err_out, "kmag: Dvojnásobný kořen?\n");
  if (err)
    *err = groot1_err;
  return mag;
}

// TODO Dodělat možnosti nastavení kořene. Větší variabilitu stop-podmínek
// Dodělat vnější nastavení přesnosti / velikosti kroku...
//#define D_HSTART 1e-5
//#define D_EPSABS 1e-5
//#define D_EPSREL 1e-5
double D_HSTART = 1e-4;
double D_EPSABS = 1e-4;
double D_EPSREL = 1e-4;
void
trace (rvec * pocatek, rvec * smer, double omega,
       int (*stop_podminky) (rvec poloha, void *param),
       void *stop_podminky_param, double max_t, double krok, susc_t susc,
       void *susc_p, int mod)
{
  rvec k = rvec_normuj (smer), r = *pocatek;
  double k_m = kmag (mod, r, k, omega, susc, susc_p, NULL);
  k = rvec_mul_con (k_m, &k);

  struct g_par p = { r, k, omega, susc, susc_p };
  gsl_odeiv2_system sys = { derivace, NULL, 6, &p };
// Zde zřejmě budu chtít něco jiného... (standard namísto y?)
  gsl_odeiv2_driver *d = gsl_odeiv2_driver_alloc_y_new
    (&sys, gsl_odeiv2_step_rkf45, D_HSTART, D_EPSABS, D_EPSREL);

  double y[6];
  for (int i = 0; i < 3; i++)
    y[i] = r.val[i];
  for (int i = 0; i < 3; i++)
    y[i + 3] = k.val[i];
  double t = 0;

  while (!stop_podminky (r, stop_podminky_param) && (t < max_t))
    {
      int status = gsl_odeiv2_driver_apply (d, &t, t + krok, y);
      if (status != GSL_SUCCESS)
	{
	  fprintf (_err_out, "trace: chyba %d\n", status);
	  break;
	}
      for (int i = 0; i < 3; i++)
	r.val[i] = y[i];
#ifdef K_NORM
      //--- Kontrola
      rvec kref = { {y[3], y[4], y[5]} };
      double kref_n = rvec_norma (&kref);
      double kref_m1 = kmag (1, r, kref, omega, susc, susc_p, NULL);
      double kref_m2 = kmag (-1, r, kref, omega, susc, susc_p, NULL);
//      fprintf (stderr, "%e %e %e %e\n", t, kref_n, kref_m1, kref_m2);
      //--- pokus o opravu
      for (int i = 0; i < 3; i++)
	y[i + 3] *= ((mod == 1) ? kref_m1 : kref_m2) / kref_n;
      //---
#endif
      printf ("%e %e %e %e %e %e %e\n", t, y[0], y[1], y[2], y[3], y[4],
	      y[5]);

    }

  gsl_odeiv2_driver_free (d);
}
